"""
Utility functions for saving and loading ei objects to and from disk
"""

import xarray as xr
import arviz as az
from pyei.two_by_two import TwoByTwoEI, TwoByTwoEIBaseBayes
from pyei.r_by_c import RowByColumnEI


def to_netcdf(ei_object, filepath):
    """
    Saves traces and some metadata for EI objects to disk

    Parameters:
    -----------
    ei_object: an object of class TwoByTwoBaseBayes or RowByColumnEI
    filepath : str
    The path to the file where the data will be saved
    Notes:
    ------

    """
    # Check if model has been fit yet
    # Add properties as attrs to sim_trace
    # if not (isinstance(ei_object, TwoByTwoEIBaseBayes) or isinstance(ei_object, RowByColumnEI)):
    #     raise ValueError("ei_object must be of class TwoByTwoEIBaseBayes or RowByColumnEI")

    is_two_by_two = isinstance(ei_object, TwoByTwoEIBaseBayes)  # bool

    if is_two_by_two:
        attr_list = [
            "model_name",
            "precinct_pops",
            "precinct_names",
            "demographic_group_name",
            "candidate_name",
            "demographic_group_fraction",
            "votes_fraction",
        ]
        ei_object.sim_trace.posterior.attrs["is_two_by_two"] = (
            "true"  # store whether 2 by 2 or r by c
        )

    else:  # r by c
        # attr_list=[]
        attr_list = [
            "model_name",
            "precinct_pops",
            "precinct_names",
            "demographic_group_names",
            "candidate_names",
            "num_groups_and_num_candidates",
        ]
        ei_object.sim_trace.posterior.attrs["is_two_by_two"] = "false"

    for attr in attr_list:
        if getattr(ei_object, attr) is not None:
            ei_object.sim_trace.posterior.attrs[attr] = getattr(ei_object, attr)

    # Use az.InferenceData's to_netcdf
    ei_object.sim_trace.to_netcdf(filepath)

    if not is_two_by_two:
        mode = "a"
        for attr in ["demographic_group_fractions", "votes_fractions"]:  # array atts
            data = xr.DataArray(getattr(ei_object, attr), name=attr)
            data.load()
            data.to_netcdf(filepath, mode=mode, group=attr, engine="netcdf4")
            data.close()


def from_netcdf(filepath):
    """
    Loads traces and metadata for EI objects to disk

    Parameters
    ----------
    filepath : str
    The path to the file from which the data will loaded
    Returns:
    --------
    ei: an object of type TwoByTwoEI or RowByColumnEI
    with sim_trace and most other atrributes set as they would
    be when fit. Note sim_model is not saved/loaded
    """
    idata = az.from_netcdf(filepath, engine="netcdf4")

    attrs_dict = idata.posterior.attrs  # pylint: disable=no-member
    attr_list = list(idata.posterior.attrs.keys())  # pylint: disable=no-member
    attr_list.remove("created_at")
    attr_list.remove("arviz_version")

    if attrs_dict["is_two_by_two"] == "true":
        ei_object = TwoByTwoEI(attrs_dict["model_name"])  # initialize EI object
        ei_object.calculate_sampled_voting_prefs()  # calculate polity-wide samples
    else:  # otherwise it's an R by C
        ei_object = RowByColumnEI(attrs_dict["model_name"])

        # pylint: disable=no-member
        ei_object.demographic_group_fractions = idata.demographic_group_fractions[
            "demographic_group_fractions"
        ].to_numpy()  # pylint: disable=no-member
        del idata.demographic_group_fractions  # pylint: disable=no-member
        # pylint: disable=no-member
        ei_object.votes_fractions = idata.votes_fractions[
            "votes_fractions"
        ].to_numpy()  # pylint: disable=no-member
        del idata.votes_fractions  # pylint: disable=no-member

    for attr in attr_list:  # set attrs of the EI object
        setattr(ei_object, attr, attrs_dict[attr])
        del idata.posterior.attrs[  # pylint: disable=no-member
            attr
        ]  # these vars only attached to the posterior for saving/loading

    ei_object.sim_trace = idata
    ei_object.calculate_summary()  # calculate summary quantities

    return ei_object



"""Test plot utils."""

import pytest
import numpy as np

from pyei import data
from pyei.plot_utils import *  # pylint:disable=wildcard-import,unused-wildcard-import
from pyei.two_by_two import TwoByTwoEI


@pytest.fixture(scope="session")
def example_two_by_two_data():
    """load santa clara data to test two by two ei and plots"""
    sc_data = data.Datasets.Santa_Clara.to_dataframe()
    group_fractions = np.array(sc_data["pct_e_asian_vote"])
    votes_fractions = np.array(sc_data["pct_for_hardy2"])
    precinct_pops = np.array(sc_data["total2"])
    demographic_group_name = "e_asian"
    candidate_name = "Hardy"
    precinct_names = sc_data["precinct"]
    return {
        "group_fractions": group_fractions,
        "votes_fractions": votes_fractions,
        "precint_pops": precinct_pops,
        "demographic_group_name": demographic_group_name,
        "candidate_name": candidate_name,
        "precinct_names": precinct_names,
    }


@pytest.fixture(scope="session")
def example_two_by_two_ei(example_two_by_two_data):  # pylint: disable=redefined-outer-name
    """run example two by two ei method - can use to test plotting"""
    ei_ex = TwoByTwoEI(model_name="king99_pareto_modification", pareto_scale=8, pareto_shape=2)
    ei_ex.fit(
        example_two_by_two_data["group_fractions"],
        example_two_by_two_data["votes_fractions"],
        example_two_by_two_data["precint_pops"],
        demographic_group_name=example_two_by_two_data["demographic_group_name"],
        candidate_name=example_two_by_two_data["candidate_name"],
        precinct_names=example_two_by_two_data["precinct_names"],
        draws=100,
        tune=100,
    )
    return ei_ex


def test_tomography_plot(example_two_by_two_data):  # pylint: disable=redefined-outer-name
    tomography_plot(
        example_two_by_two_data["group_fractions"],
        example_two_by_two_data["votes_fractions"],
        example_two_by_two_data["demographic_group_name"],
        example_two_by_two_data["candidate_name"],
    )


def test_ei_plot_and_plot_summary(example_two_by_two_ei):  # pylint: disable=redefined-outer-name
    # TODO: maybe uncouple this to test the plot utils piece alone
    axes = example_two_by_two_ei.plot()
    assert len(axes) == 2  ## one axis each for boxplot and kde


def test_ei_plot_kdes(example_two_by_two_ei):  # pylint: disable=redefined-outer-name
    # TODO: maybe uncouple this to test the plot utils piece alone
    ax = example_two_by_two_ei.plot_kde()
    assert ax is not None


def test_ei_plot_boxplot(example_two_by_two_ei):  # pylint: disable=redefined-outer-name
    # TODO: maybe uncouple this to test the plot utils piece alone
    ax = example_two_by_two_ei.plot_boxplot()
    assert ax is not None


def test_ei_plot_intervals(example_two_by_two_ei):  # pylint: disable=redefined-outer-name
    # TODO: maybe uncouple this to test the plot utils piece alone
    ax = example_two_by_two_ei.plot_intervals()
    assert ax is not None


def test_ei_precinct_level_plots(example_two_by_two_ei):  # pylint: disable=redefined-outer-name
    # TODO: maybe uncouple this to test the plot utils piece alone
    ax = example_two_by_two_ei.precinct_level_plot()
    assert ax is not None


def test_ei_plot_intervals_by_precinct(
    example_two_by_two_ei,
):  # pylint: disable=redefined-outer-name
    # TODO: maybe uncouple this to test the plot utils piece alone
    ax = example_two_by_two_ei.plot_intervals_by_precinct()
    assert ax is not None


def test_plot_polarization_kde(example_two_by_two_ei):  # pylint: disable=redefined-outer-name
    percentile_ax = example_two_by_two_ei.plot_polarization_kde(threshold=0.4, show_threshold=True)
    percentile_ax_2 = example_two_by_two_ei.plot_polarization_kde(
        threshold=0.4, reference_group=1, show_threshold=True
    )
    threshold_ax = example_two_by_two_ei.plot_polarization_kde(percentile=95, show_threshold=True)
    assert percentile_ax is not None
    assert percentile_ax_2 is not None
    assert threshold_ax is not None
    with pytest.raises(ValueError):
        example_two_by_two_ei.plot_polarization_kde(
            threshold=0.4, percentile=95, show_threshold=True
        )



"""Test Greiner Quinn Gibbs sampler."""

# pylint: disable=duplicate-code
import random
import pytest
import numpy as np
import scipy.stats as st


from pyei import data
from pyei.r_by_c import RowByColumnEI
from pyei.greiner_quinn_gibbs_sampling import (
    get_initial_internal_count_sample,
    theta_to_omega,
    greiner_quinn_gibbs_sample,
)
from pyei.distribution_utils import non_central_hypergeometric_sample


@pytest.fixture(scope="session")
def example_r_by_c_data_asym():
    """trimmed santa clara dataset with r not equal to c"""
    sc_data = data.Datasets.Santa_Clara.to_dataframe()
    sc_data = sc_data.iloc[:10, :]
    precinct_pops = np.array(sc_data["total2"])
    votes_fractions = np.array(sc_data[["pct_for_hardy2", "pct_for_kolstad2", "pct_for_nadeem2"]]).T
    candidate_names = ["Hardy", "Kolstad", "Nadeem"]
    group_fractions = np.array(sc_data[["pct_asian_vote", "pct_non_asian_vote"]]).T
    demographic_group_names = ["asian", "non_asian"]
    group_counts = np.round(group_fractions * precinct_pops)
    vote_counts = np.round(votes_fractions * precinct_pops)
    num_groups = group_counts.shape[0]
    num_candidates = vote_counts.shape[0]

    group_diff = group_counts.sum(axis=0) - precinct_pops
    for idx_of_mismatch in np.where(group_diff != 0):
        group_to_adjust = random.randint(0, num_groups - 1)
        group_counts[group_to_adjust, idx_of_mismatch] -= group_diff[idx_of_mismatch]

    vote_diff = vote_counts.sum(axis=0) - precinct_pops
    for idx_of_mismatch in np.where(vote_diff != 0):
        candidate_to_adjust = random.randint(0, num_candidates - 1)
        vote_counts[candidate_to_adjust, idx_of_mismatch] -= vote_diff[idx_of_mismatch]

    group_counts = group_counts.T
    vote_counts = vote_counts.T
    return {
        "group_fractions": group_fractions,
        "votes_fractions": votes_fractions,
        "group_counts": group_counts,
        "vote_counts": vote_counts,
        "precinct_pops": precinct_pops,
        "demographic_group_names": demographic_group_names,
        "candidate_names": candidate_names,
    }


def test_get_initial_internal_count_sample(
    example_r_by_c_data_asym,
):  # pylint: disable=redefined-outer-name:
    vote_counts = example_r_by_c_data_asym["vote_counts"]
    group_counts = example_r_by_c_data_asym["group_counts"]
    precinct_pops = example_r_by_c_data_asym["precinct_pops"]
    samp = get_initial_internal_count_sample(group_counts, vote_counts, precinct_pops)
    samp_py = get_initial_internal_count_sample.py_func(group_counts, vote_counts, precinct_pops)

    assert np.all(samp.sum(axis=2) - group_counts == 0)  # sample respects given group counts
    assert np.all(samp.sum(axis=1) - vote_counts == 0)  # sample respects given vote counts
    assert np.all(samp_py.sum(axis=2) - group_counts == 0)  # sample respects given group counts
    assert np.all(samp_py.sum(axis=1) - vote_counts == 0)  # sample respects given vote counts


def test_theta_to_omega():
    num_precincts = 8
    r = 3
    c = 4
    alpha = np.ones(c)
    theta = st.dirichlet.rvs(alpha, size=(num_precincts, r))
    omega = theta_to_omega(theta)
    assert omega.shape[2] == c - 1
    np.testing.assert_almost_equal(
        omega[3, 2, 1], np.log(theta[3, 2, 1] / theta[3, 2, c - 1]), decimal=4
    )


def test_greiner_quinn_gibbs_sample(
    example_r_by_c_data_asym,
):  # pylint: disable=redefined-outer-name
    r = example_r_by_c_data_asym["group_counts"].shape[1]
    c = example_r_by_c_data_asym["vote_counts"].shape[1]
    print(r, c)
    num_samples = 100
    nu_0 = 10
    psi_0 = 1 / 10 * np.identity(r * (c - 1))  # relates to prior precision
    k_0_inv = 1 / (0.5) * np.identity(r * (c - 1))  # same size as Sigma

    votes_all_precincts = example_r_by_c_data_asym["vote_counts"].sum(axis=0)
    log_ratio_support = np.log(
        votes_all_precincts[0:-1] / votes_all_precincts[-1]
    )  # assume each group supports according to the total vote split
    mu_0 = np.tile(log_ratio_support, r)  # shape r * (c-1)
    gamma = 0.1

    greiner_quinn_gibbs_sample(
        example_r_by_c_data_asym["group_counts"],
        example_r_by_c_data_asym["vote_counts"],
        num_samples,
        nu_0,
        psi_0,
        k_0_inv,
        mu_0,
        gamma=gamma,
    )


def test_pyei_greiner_quinn_gibbs(
    example_r_by_c_data_asym,
):  # pylint: disable=redefined-outer-name
    ei_greiner_quinn = RowByColumnEI(model_name="greiner-quinn")
    ei_greiner_quinn.fit(
        example_r_by_c_data_asym["group_fractions"],
        example_r_by_c_data_asym["votes_fractions"],
        example_r_by_c_data_asym["precinct_pops"],
        num_samples=5,
        burnin=1,
    )


def test_non_central_hypergeometric_sample():
    samp = non_central_hypergeometric_sample.py_func(10, 5, 7, 1)
    assert samp >= 2
    assert samp <= 10
    samp2 = non_central_hypergeometric_sample.py_func(10, 10, 7, 1)
    assert samp2 <= 10



"""
Models and fitting for 2x2 methods

"""

import warnings
import pymc as pm
import numpy as np
import pytensor.tensor as at
import pytensor
from .plot_utils import (
    plot_conf_or_credible_interval,
    plot_boxplots,
    plot_kdes,
    plot_precincts,
    plot_polarization_kde,
    plot_summary,
    plot_intervals_all_precincts,
)

__all__ = ["TwoByTwoEI", "ei_beta_binom_model_modified"]


def truncated_normal_asym(
    group_fraction, votes_fraction, precinct_pops
):  # pylint: disable=too-many-locals
    """
    A modification of king97's truncated normal that puts some broad priors
    over the parameters of the truncated normal dist

    Parameters
    ----------
    group_fraction: Length-p (p=# of precincts) vector giving demographic information
        as the fraction of precinct_pop in the demographic group of interest
    votes_fraction: Length p vector giving the fraction of each precinct_pop that
        votes for the candidate of interest
    precinct_pops: Length-p vector giving size of each precinct population of interest
         (e.g. voting population)

    Returns
    -------
    model: A pymc3 model

    """
    num_precincts = len(precinct_pops)
    b_1_l_bound = np.maximum(0, (votes_fraction - 1 + group_fraction) / group_fraction)
    b_1_u_bound = np.minimum(1, votes_fraction / group_fraction)
    b_2_l_bound = np.maximum(0, (votes_fraction - group_fraction) / (1 - group_fraction))
    b_2_u_bound = np.minimum(1, (votes_fraction) / (1 - group_fraction))

    # For stability, use whichever of b_1 and b_2 has the broader width between bounds as
    # the parameter that we include in the model (then the other of b_1 or b_2 will
    # be calculated deterministically from that and votes_fraction) -- the deterministically
    # calculated one is here marked as "lower-level" one
    if (b_2_u_bound - b_2_l_bound).mean() > (b_1_u_bound - b_1_l_bound).mean():  # swap
        group_fraction = 1 - group_fraction
        upper_level_b_name = "b_2"
        lower_level_b_name = "b_1"
        upper_level_tn_mean_name = "tn_mean_2"
        lower_level_tn_mean_name = "tn_mean_1"
        upper_level_sigma_name = "sigma_22"
        lower_level_sigma_name = "sigma_11"
        upper_level_l_bound = b_2_l_bound
        upper_level_u_bound = b_2_u_bound
    else:  # not swapping
        upper_level_b_name = "b_1"
        lower_level_b_name = "b_2"
        upper_level_tn_mean_name = "tn_mean_1"
        lower_level_tn_mean_name = "tn_mean_2"
        upper_level_sigma_name = "sigma_11"
        lower_level_sigma_name = "sigma_22"
        upper_level_l_bound = b_1_l_bound
        upper_level_u_bound = b_1_u_bound

    with pm.Model() as model:
        sigma_upper = pm.HalfNormal(
            upper_level_sigma_name, sigma=0.707
        )  # chosen to match King 97 #sigma11
        sigma_lower = pm.HalfNormal(
            lower_level_sigma_name, sigma=0.707
        )  # chosen to match king 97 #sigma22
        rho = pm.Uniform("rho", -0.5, 0.5)  # TODO: revisit
        sigma_12 = sigma_upper * sigma_lower * rho

        tn_mean_upper = pm.Uniform(upper_level_tn_mean_name)
        tn_mean_lower = pm.Uniform(lower_level_tn_mean_name)

        upper_b = pm.TruncatedNormal(
            upper_level_b_name,
            mu=tn_mean_upper,
            sigma=sigma_upper,
            shape=num_precincts,
            lower=upper_level_l_bound,
            upper=upper_level_u_bound,
        )

        mu_i = tn_mean_upper * group_fraction + tn_mean_lower * (1 - group_fraction)
        w_i = pm.Deterministic(
            "w_i", sigma_upper**2 * group_fraction + sigma_12 * (1 - group_fraction)
        )
        sigma_i_sq = pm.Deterministic(
            "sigma_i_sq",
            sigma_lower**2
            + 2 * (sigma_12 - sigma_lower**2) * group_fraction
            + (sigma_upper * 2 + sigma_lower**2 - 2 * sigma_12) * group_fraction**2,
        )

        votes_frac_mean = mu_i + w_i * (upper_b - tn_mean_upper) / (sigma_upper**2)
        votes_frac_var = pm.Deterministic("votes_frac_var", sigma_i_sq - w_i**2 / (sigma_upper**2))

        votes_frac_l_bound = group_fraction * upper_b
        votes_frac_u_bound = (1 - group_fraction) + group_fraction * upper_b

        votes_frac_stdev = pm.math.sqrt(votes_frac_var)
        pm.TruncatedNormal(
            "votes_fraction",
            mu=votes_frac_mean,
            sigma=votes_frac_stdev,
            lower=votes_frac_l_bound,
            upper=votes_frac_u_bound,
            observed=votes_fraction,
        )
        pm.Deterministic(
            lower_level_b_name,
            (votes_fraction - upper_b * group_fraction) / (1 - group_fraction),
        )
    return model


def ei_beta_binom_model_modified(
    group_fraction, votes_fraction, precinct_pops, pareto_scale=8, pareto_shape=2
):
    """
    An modification of the 2 x 2 beta/binomial EI model from King, Rosen, Tanner 1999,
    with (scaled) Pareto distributions over each parameters of the beta distribution,
    for better sampling geometry

    Parameters
    ----------
    group_fraction: Length-p (p=# of precincts) vector giving demographic information (X)
        as the fraction of precinct_pop in the demographic group of interest
    votes_fraction: Length p vector giving the fraction of each precinct_pop that votes
        for the candidate of interest (T)
    precinct_pops: Length-p vector giving size of each precinct population of interest
        (e.g. voting population) (N)
    pareto_scale: A positive real number. The scale paremeter passed to the
        pareto hyperparamters
    pareto_shape: A positive real number. The shape paremeter passed to the
        pareto hyperparamters

    Returns
    -------
    model: A pymc3 model

    Notes
    -----
    Reparametrizing of the hyperpriors to give (hopefully) better geometry for sampling.
    Also gives intuitive interpretation of hyperparams as mean and counts
    """
    votes_count_obs = votes_fraction * precinct_pops
    num_precincts = len(precinct_pops)
    # tot_pop = precinct_pops.sum()
    with pm.Model() as model:
        phi_1 = pm.Uniform("phi_1", lower=0.0, upper=1.0)
        kappa_1 = pm.Pareto("kappa_1", m=pareto_scale, alpha=pareto_shape)

        phi_2 = pm.Uniform("phi_2", lower=0.0, upper=1.0)
        kappa_2 = pm.Pareto("kappa_2", m=pareto_scale, alpha=pareto_shape)

        b_1 = pm.Beta(
            "b_1",
            alpha=phi_1 * kappa_1,
            beta=(1.0 - phi_1) * kappa_1,
            shape=num_precincts,
        )
        b_2 = pm.Beta(
            "b_2",
            alpha=phi_2 * kappa_2,
            beta=(1.0 - phi_2) * kappa_2,
            shape=num_precincts,
        )

        theta = group_fraction * b_1 + (1 - group_fraction) * b_2
        pm.Binomial("votes_count", n=precinct_pops, p=theta, observed=votes_count_obs)

    return model


def ei_beta_binom_model(group_fraction, votes_fraction, precinct_pops, lmbda):
    """
    2 x 2 beta/binomial EI model from King, Rosen, Tanner 1999

    Parameters
    ----------
    group_fraction: Length-p (p=# of precincts) vector giving demographic information
        as the fraction of precinct_pop in the demographic group of interest
    votes_fraction: Length p vector giving the fraction of each precinct_pop that
        votes for the candidate of interest
    precinct_pops: Length-p vector giving size of each precinct population of interest
         (e.g. voting population)
    lmbda: a hyperparameter governing the exponential distributions
        over the parameters for the parameters of the beta distributions
        King sets this to 0.5, but lower seems to produce better geometry
        for sampling

    Returns
    -------
    model: A pymc3 model
    """

    votes_count_obs = votes_fraction * precinct_pops
    num_precincts = len(precinct_pops)
    with pm.Model() as model:
        c_1 = pm.Exponential("c_1", lmbda)
        d_1 = pm.Exponential("d_1", lmbda)
        c_2 = pm.Exponential("c_2", lmbda)
        d_2 = pm.Exponential("d_2", lmbda)

        b_1 = pm.Beta("b_1", alpha=c_1, beta=d_1, shape=num_precincts)
        b_2 = pm.Beta("b_2", alpha=c_2, beta=d_2, shape=num_precincts)

        theta = group_fraction * b_1 + (1 - group_fraction) * b_2
        pm.Binomial("votes_count", n=precinct_pops, p=theta, observed=votes_count_obs)
    return model


def log_binom_sum(lower, upper, obs_vote, n0_curr, n1_curr, b_1_curr, b_2_curr, prev):
    """
    Helper function for computing log prob of convolution of binomial

    Parameters
    ----------
    lower, upper : lower and upper bounds on the (unobserved) count of votes from given
        deographic group for given candidate within precinct
    n0_curr: the (current value for the) count of given demographic group in the precinct
    n1_curr: the (current value for the)count of the complement of given demographic group
         in the precinct
    b_1_curr: corresponds to p0 in wakefield's notation, the probability that an individual
        in the given demographic group votes for the given candidate
    b_2_curr: corresponds to p1 in wakefield's notation, the probability that an individual
        in the complement of the given demographic group votes for the given candidate
    prev: the partial sum of the log prob of the convolution of binomials

    Returns
    -------
    prev + component_for_current_precinct : sum of the previous value of the partial sum
        for the log prob of the convolution of binomials and an additional term for one
        precinct

    """

    # votes_within_group_count is y_0i in Wakefield's notation, the count of votes from
    # given group for given candidate within precinct i (unobserved)
    votes_within_group_count = at.arange(lower, upper)
    component_for_current_precinct = pm.math.logsumexp(
        pm.logp(pm.Binomial.dist(n0_curr, b_1_curr), votes_within_group_count)
        + pm.logp(pm.Binomial.dist(n1_curr, b_2_curr), obs_vote - votes_within_group_count)
    )
    return prev + component_for_current_precinct


def binom_conv_log_p(b_1, b_2, n_0, n_1, upper, lower, obs_votes):
    """
    Log probability for convolution of binomials

    Parameters
    ----------
    b_1: corresponds to p0 in wakefield's notation, the probability that an individual
        in the given demographic group votes for the given candidate
    b_2: corresponds to p1 in wakefield's notation, the probability that an individual
        in the complement of the given demographic group votes for the given candidate

    n_0: the count of given demographic group in the precinct
    n_1: the count of the complement of given demographic group in the precinct

    lower, upper : lower and upper bounds on the (unobserved) count of votes from given
    deographic group for given candidate within precinct
    (corresponds to votes_within_group_count in log_binom_sum and y_0i in Wakefield's)

    Returns
    -------
    A theano tensor giving the log probability of b_1, b_2 (given the other parameters)

    Notes
    -----
    See Wakefield 2004 equation 4
    """

    result, _ = pytensor.scan(
        fn=log_binom_sum,
        outputs_info={"taps": [-1], "initial": at.as_tensor(np.array([0.0]))},
        sequences=[
            at.as_tensor(lower),
            at.as_tensor(upper),
            at.as_tensor(obs_votes),
            at.as_tensor(n_0),
            at.as_tensor(n_1),
            at.as_tensor(b_1),
            at.as_tensor(b_2),
        ],
    )
    return result[-1]


# def wakefield_model_beta(
#     group_fraction, votes_fraction, precinct_pops, pareto_scale=8, pareto_shape=2
# ):
#     """
#     2 x 2 EI model based on Wakefield's, with pareto distributions in upper level of hierarchy

#     Parameters
#     ----------
#     group_fraction: Length-p (p=# of precincts) vector giving demographic information
#         as the fraction of precinct_pop in the demographic group of interest
#     votes_fraction: Length p vector giving the fraction of each precinct_pop that
#         votes for the candidate of interest
#     precinct_pops: Length-p vector giving size of each precinct population of interest
#          (e.g. voting population)

#     Returns
#     -------
#     model: A pymc3 model
#     """

#     vote_count_obs = votes_fraction * precinct_pops
#     group_count_obs = group_fraction * precinct_pops
#     num_precincts = len(precinct_pops)
#     upper = np.minimum(group_count_obs, vote_count_obs)  # upper bound on y
#     lower = np.maximum(0.0, vote_count_obs - precinct_pops + group_count_obs)  # lower bound on y
#     with pm.Model() as model:
#         phi_1 = pm.Uniform("phi_1", lower=0.0, upper=1.0)
#         kappa_1 = pm.Pareto("kappa_1", m=pareto_scale, alpha=pareto_shape)

#         phi_2 = pm.Uniform("phi_2", lower=0.0, upper=1.0)
#         kappa_2 = pm.Pareto("kappa_2", m=pareto_scale, alpha=pareto_shape)

#         b_1 = pm.Beta(
#             "b_1",
#             alpha=phi_1 * kappa_1,
#             beta=(1.0 - phi_1) * kappa_1,
#             shape=num_precincts,
#         )
#         b_2 = pm.Beta(
#             "b_2",
#             alpha=phi_2 * kappa_2,
#             beta=(1.0 - phi_2) * kappa_2,
#             shape=num_precincts,
#         )

#         pm.DensityDist(
#             "votes_count_obs",
#             b_1,
#             b_2,
#             group_count_obs,
#             precinct_pops - group_count_obs,
#             upper,
#             lower,
#             observed=vote_count_obs,
#             logp=binom_conv_log_p,
#         )
#     return model


# def wakefield_normal(group_fraction, votes_fraction, precinct_pops, mu0=0, mu1=0):
#     """
#     2 x 2 EI model Wakefield with normal hyperpriors

#     Note: Wakefield suggests adding another level of hierarchy, with a prior over mu0 and mu1,
#     sigma0, sigma1, but that is not yet implemented here

#     Parameters
#     ----------
#     group_fraction: Length-p (p=# of precincts) vector giving demographic information
#         as the fraction of precinct_pop in the demographic group of interest
#     votes_fraction: Length p vector giving the fraction of each precinct_pop that
#         votes for the candidate of interest
#     precinct_pops: Length-p vector giving size of each precinct population of interest
#          (e.g. voting population)
#     mu0: float
#         Mean of the normally distributed hyperparameter associated with the demographic
#         group of interest
#     m1: Mean of the normally distributed hyperparameter associated with the complement
#     of the demographic group of interest

#     Returns
#     -------
#     model: A pymc3 model

#     """

#     vote_count_obs = votes_fraction * precinct_pops
#     group_count_obs = group_fraction * precinct_pops
#     num_precincts = len(precinct_pops)
#     upper = np.minimum(group_count_obs, vote_count_obs)  # upper bound on y
#     lower = np.maximum(0.0, vote_count_obs - precinct_pops + group_count_obs)  # lower bound on y
#     with pm.Model() as model:
#         sigma_0 = pm.Gamma("sigma0", 1, 0.1)
#         sigma_1 = pm.Gamma("sigma1", 1, 0.1)

#         theta_0 = pm.Normal("theta0", mu0, sigma_0, shape=num_precincts)
#         theta_1 = pm.Normal("theta1", mu1, sigma_1, shape=num_precincts)

#         b_1 = pm.Deterministic(
#             "b_1", at.exp(theta_0) / (1 + at.exp(theta_0))
#         )  # vector of length num_precincts
#         b_2 = pm.Deterministic(
#             "b_2", at.exp(theta_1) / (1 + at.exp(theta_1))
#         )  # vector of length num_precincts

#         pm.DensityDist(
#             "votes_count_obs",
#             b_1,
#             b_2,
#             group_count_obs,  # n_0
#             precinct_pops - group_count_obs,  # n_1
#             upper,
#             lower,
#             observed=vote_count_obs,  # obs_votes
#             logp=binom_conv_log_p,
#         )
#     return model


class TwoByTwoEIBaseBayes:
    """
    Init, summary, plots for 2 x 2 EI models that proceed via sampling

    Note: does not assume precinct level samples are available,
    because this is not possible for goodman_er_bayes. So, plots of precinct
    level quanitities are defined in the subclass TwoByTwoEi

    Note: subclass will need to define methods to fit (sample), and
        to define summary quantities
    """

    def __init__(self, model_name, **additional_model_params):
        """
        model_name: str
            The name of one of the models ( "king99", "king99_pareto_modification",
             "truncated_normal",
            "goodman_er_bayes")
        additional_model_params
            Hyperparameters to pass to model, if changing default parameters
            (see model documentation for the hyperparameters for each model)
        """
        self.model_name = model_name
        self.additional_model_params = additional_model_params

        self.sim_model = None
        self.sim_trace = None

        self.precinct_pops = None
        self.demographic_group_name = None
        self.candidate_name = None

        self.demographic_group_fraction = None
        self.votes_fraction = None

        self.posterior_mean_voting_prefs = [None, None]
        self.credible_interval_95_mean_voting_prefs = [None, None]
        self.sampled_voting_prefs = [None, None]

    def group_names_for_display(self):
        """Returns the group names to be displayed in plots"""
        return self.demographic_group_name, "non-" + self.demographic_group_name

    def _voting_prefs_array(self):
        """Bundles together the samples as num_samples x 2 x 1 array,
        for ease of passing to plots"""
        num_samples = len(self.sampled_voting_prefs[0])
        sampled_voting_prefs = np.empty((num_samples, 2, 1))  # num_samples x 2 x 1
        sampled_voting_prefs[:, 0, 0] = self.sampled_voting_prefs[0]
        sampled_voting_prefs[:, 1, 0] = self.sampled_voting_prefs[1]
        return sampled_voting_prefs

    def calculate_summary(self):
        """Calculate point estimates (post. means) and 95% equal-tailed credible intervals
        Assumes sampled_voting_prefs has already been set"""

        # compute point estimates
        self.posterior_mean_voting_prefs[0] = self.sampled_voting_prefs[0].mean()
        self.posterior_mean_voting_prefs[1] = self.sampled_voting_prefs[1].mean()

        # compute credible intervals
        percentiles = [2.5, 97.5]
        self.credible_interval_95_mean_voting_prefs[0] = np.percentile(
            self.sampled_voting_prefs[0], percentiles
        )
        self.credible_interval_95_mean_voting_prefs[1] = np.percentile(
            self.sampled_voting_prefs[1], percentiles
        )

    def _calculate_polarization(self, threshold=None, percentile=None, reference_group=0):
        """calculate percentile given a threshold, or threshold if given a percentile
        exactly one of percentile and threshold must be null

        Parameters
        ----------
        threshold: float
            If not None, function will return the estimated probability that difference
            between the two groups' preferences for the given candidate is more than
            {threshold}
        percentile: float
            If not None, function will return the threshold for which {percentile}
            equals the estimated probability that difference between the two groups'
            preferences for the given candidate is more than {threshold}
        reference group: int {0, 1}
            The index of the reference group. If 0, the thresholds are calcuated as
            (group 0 preferences - group 1 preferences). If 1, the thresholds are
            calculated as (group 1 preferences - group 0 preferences)

        Notes
        -----
        Exactly one of threshold and percentile must be None

        """
        samples = self.sampled_voting_prefs[0] - self.sampled_voting_prefs[1]
        group = self.demographic_group_name
        group_complement = "non-" + self.demographic_group_name
        if reference_group == 1:
            samples = -samples
            group = "non-" + self.demographic_group_name
            group_complement = self.demographic_group_name

        if percentile is None and threshold is not None:
            percentile = 100 * (samples > threshold).sum() / len(self.sampled_voting_prefs[0])
        elif threshold is None and percentile is not None:
            threshold = np.percentile(samples, 100 - percentile)
        else:
            raise ValueError(
                """Exactly one of threshold or percentile must be None.
            Set a threshold to calculate the associated percentile, or a percentile
            to calculate the associated threshold.
            """
            )
        return threshold, percentile, samples, [group, group_complement]

    def polarization_report(self, threshold=None, percentile=None, reference_group=0, verbose=True):
        """
        For a given threshold, return the probability that difference between the group's
        preferences for the given candidate is more than threshold
        OR
        For a given confidence level, return the associated central credible interval for
        the difference between the two groups' preferences.
        Exactly one {percentile,threshold} must be None

        Parameters
        ----------
        threshold: float
            If not None, function will return the estimated probability that difference
            between the two groups' preferences for the given candidate is more than
            {threshold}
        percentile: float
            If not None, function will return the central interval which {percentile}
            equals the estimated probability that difference between the two groups'
            preferences for the given candidate is in that interval
        reference group: int {0, 1}
            The index of the reference group. If 0, the thresholds are calcuated as
            (group 0 preferences - group 1 preferences). If 1, the thresholds are
            calculated as (group 1 preferences - group 0 preferences)
        verbose: bool
            If True, print out a report

        Notes
        -----
        Exactly one of threshold and percentile must be None
        """
        return_interval = threshold is None

        if return_interval:
            lower_percentile = (100 - percentile) / 2
            upper_percentile = lower_percentile + percentile
            lower_threshold, _, _, groups = self._calculate_polarization(
                threshold, upper_percentile, reference_group
            )
            upper_threshold, _, _, groups = self._calculate_polarization(
                threshold, lower_percentile, reference_group
            )
            if verbose:
                print(
                    f"There is a {percentile}% probability that the difference between the groups'"
                    + f" preferences for {self.candidate_name} ({groups[0]} - {groups[1]}) is "
                    + f"between [{lower_threshold:.2f}, {upper_threshold:.2f}]."
                )
            return (lower_threshold, upper_threshold)
        else:
            threshold, percentile, _, groups = self._calculate_polarization(
                threshold, percentile, reference_group
            )
            if verbose:
                print(
                    f"There is a {percentile:.1f}% probability that the difference between"
                    + f" the groups' preferences for {self.candidate_name} ({groups[0]} -"
                    + f"  {groups[1]}) is more than {threshold:.2f}."
                )
            return percentile

    def summary(self):
        """Return a summary string"""
        # TODO: probably format this as a table
        return f"""Model: {self.model_name}
        Computed from the raw b_i samples by multiplying by population and then getting
        the proportion of the total pop (total pop=summed across all districts):
        The posterior mean for the district-level voting preference of
        {self.demographic_group_name} for {self.candidate_name} is
        {self.posterior_mean_voting_prefs[0]:.3f}
        The posterior mean for the district-level voting preference of
        non-{self.demographic_group_name} for {self.candidate_name} is
        {self.posterior_mean_voting_prefs[1]:.3f}
        95% equal-tailed Bayesian credible interval for district-level voting preference of
        {self.demographic_group_name} for {self.candidate_name} is
        {self.credible_interval_95_mean_voting_prefs[0]}
        95% equal-tailed Bayesian credible interval for district-level voting preference of
        non-{self.demographic_group_name} for {self.candidate_name} is
        {self.credible_interval_95_mean_voting_prefs[1]}
        """

    def plot_kde(self, ax=None):
        """kernel density estimate/ histogram plot
        Optional arguments:
        ax  :  matplotlib axes object
        """
        return plot_kdes(
            self._voting_prefs_array(),
            self.group_names_for_display(),
            [self.candidate_name],
            plot_by="candidate",
            axes=ax,
        )

    def plot_boxplot(self, ax=None):
        """Boxplot of voting prefs for each group
        Optional arguments:
        ax  :  matplotlib axes object"""
        return plot_boxplots(
            self._voting_prefs_array(),
            self.group_names_for_display(),
            [self.candidate_name],
            plot_by="candidate",
            axes=ax,
        )

    def plot_intervals(self, ax=None):
        """Plot of credible intervals for each group
        Optional arguments:
        ax  :  matplotlib axes object
        """
        title = "95% credible intervals"
        return plot_conf_or_credible_interval(
            [
                self.credible_interval_95_mean_voting_prefs[0],
                self.credible_interval_95_mean_voting_prefs[1],
            ],
            self.group_names_for_display(),
            self.candidate_name,
            title,
            ax=ax,
        )

    def plot_polarization_kde(
        self,
        threshold=None,
        percentile=None,
        reference_group=0,
        show_threshold=False,
        ax=None,
        color="steelblue",
    ):
        """
        Plot kde of differences between voting preferences

        Parameters
        ----------
        threshold: float
            If not None, function will return the estimated probability that difference
            between the two groups' preferences for the given candidate is more than
            {threshold}
        percentile: float
            If not None, function will return the central interval which {percentile}
            equals the estimated probability that difference between the two groups'
            preferences for the given candidate is in that interval
        reference group: int {0, 1}
            The index of the reference group. If 0, the thresholds are calcuated as
            (group 0 preferences - group 1 preferences). If 1, the thresholds are
            calculated as (group 1 preferences - group 0 preferences)
        show_threshold: bool (optional)
            Default: False. If true, add a vertical line at the threshold on the plot
            and display the associated tail probability
        ax: matplotlib axis object (optional)
        color : str (optional)
            Specifies a color for matplotlib to be used in the histogram/kde.
            default="steelblue"

        Returns
        -------
        Matplotlib axis object

        """
        return_interval = threshold is None

        if return_interval:
            lower_percentile = (100 - percentile) / 2
            upper_percentile = lower_percentile + percentile
            lower_threshold, _, samples, groups = self._calculate_polarization(
                threshold, upper_percentile, reference_group
            )
            upper_threshold, _, samples, groups = self._calculate_polarization(
                threshold, lower_percentile, reference_group
            )
            thresholds = [lower_threshold, upper_threshold]
        else:
            threshold, percentile, samples, groups = self._calculate_polarization(
                threshold, percentile, reference_group
            )
            thresholds = [threshold]  # pylint: disable=duplicate-code

        return plot_polarization_kde(
            samples,
            thresholds,
            percentile,
            groups,  # pylint: disable=duplicate-code
            self.candidate_name,
            show_threshold,
            ax,
            color=color,
        )


class TwoByTwoEI(TwoByTwoEIBaseBayes):
    """
    Fitting and plotting for king97, king99, and wakefield models
    """

    def __init__(self, model_name, **additional_model_params):
        """
        model_name: str
            Name of model: can be 'king97', 'king99', 'king99_pareto_modification'
            'wakefield_beta' or 'wakefield normal'
        additional_model_params
            Hyperparameters to pass to model, if changing default parameters
            (see model documentation for the hyperparameters for each model)
        """
        super().__init__(model_name, **additional_model_params)

        self.precinct_pops = None
        self.precinct_names = None

    def fit(
        self,
        group_fraction,
        votes_fraction,
        precinct_pops,
        demographic_group_name="given demographic group",
        candidate_name="given candidate",
        precinct_names=None,
        target_accept=0.99,
        tune=1500,
        draw_samples=True,
        **other_sampling_args,
    ):
        """Fit the specified model using MCMC sampling
        Required arguments:
        group_fraction  :   Length-p (p=# of precincts) vector giving demographic
                            information (X) as the fraction of precinct_pop in
                            the demographic group of interest
        votes_fraction  :   Length p vector giving the fraction of each precinct_pop
                            that votes for the candidate of interest (T)
        precinct_pops   :   Length-p vector giving size of each precinct population
                            of interest (e.g. voting population) (N)
        Optional arguments:
        demographic_group_name  :   Name of the demographic group of interest,
                                    where results are computed for the
                                    demographic group and its complement
        candidate_name          :   Name of the candidate whose support we
                                    want to analyze
        precinct_names          :   Length p vector giving the string names
                                    for each precinct.
        target_accept : float, optional
            Default=.99 Strictly between zero and 1 (should be close to 1). Passed to pymc's
            sampling.sample
        tune : int, optional
            Default=1500, passed to pymc's sampling.sample
        draw_samples: bool, optional
            Default=True. Set to False to only set up the variable but not generate
            posterior samples (i.e. if you want to generate prior predictive samples only)
        other_sampling_args :
            For to pymc's sampling.sample
            https://docs.pymc.io/api/inference.html
        """
        # Additional params includes lambda for king99, the
        # parameter passed to the exponential hyperpriors,
        # and the paretoo_scale and pareto_shape parameters for the pareto
        # dist in the king99_pareto_modification model hyperprior

        # check that lengths of group_fraction, votes_fraction, precinct_pops match
        if not (
            len(group_fraction) == len(votes_fraction) and len(votes_fraction) == len(precinct_pops)
        ):
            raise ValueError(
                """Mismatching num_precincts in inputs. \n
            votes_fraction, group_fraction, precinct_pops should all have same length
            """
            )
        self.demographic_group_fraction = group_fraction
        self.votes_fraction = votes_fraction
        self.precinct_pops = precinct_pops
        self.demographic_group_name = demographic_group_name
        self.candidate_name = candidate_name
        if precinct_names is not None:
            assert len(precinct_names) == len(precinct_pops)
            if len(set(precinct_names)) != len(precinct_names):
                warnings.warn(
                    "Precinct names are not unique. This may interfere with "
                    "passing precinct names to precinct_level_plot()."
                )
            self.precinct_names = np.array(precinct_names)

        if self.model_name == "king99":
            model_function = ei_beta_binom_model

        elif self.model_name == "king99_pareto_modification":
            model_function = ei_beta_binom_model_modified

        elif self.model_name == "truncated_normal":
            model_function = truncated_normal_asym

        self.sim_model = model_function(  # pylint: disable=possibly-used-before-assignment
            group_fraction,
            votes_fraction,
            precinct_pops,
            **self.additional_model_params,
        )

        if draw_samples:
            with self.sim_model:  # pylint: disable=not-context-manager
                # this "if" is a workaround until jax.scipy.special.erfcx is
                # implemented https://github.com/google/jax/issues/1987
                # (when that's implemented, trunc-normal can use jax sampling
                # as well) @TODO: check on this in a little while
                if self.model_name in [
                    "truncated_normal"
                ]:  # , "wakefield_normal", "wakefield_beta"]:
                    self.sim_trace = pm.sample(
                        target_accept=target_accept,
                        tune=tune,
                        **other_sampling_args,
                    )
                else:
                    self.sim_trace = pm.sample(
                        target_accept=target_accept,
                        tune=tune,
                        nuts_sampler="numpyro",
                        **other_sampling_args,
                    )

            self.calculate_sampled_voting_prefs()
            super().calculate_summary()

    def calculate_sampled_voting_prefs(self):
        """Sampled voting preferences (combining samples with precinct pops)"""
        # multiply sample proportions by precinct pops to get samples of
        # number of voters the demographic group who voted for the candidate
        # in each precinct
        samples_converted_to_pops_gp1 = (
            self.sim_trace["posterior"]["b_1"].stack(all_draws=["chain", "draw"]).values.T
            * self.precinct_pops
        )  # shape: num_samples x num_precincts
        samples_converted_to_pops_gp2 = (
            self.sim_trace["posterior"]["b_2"].stack(all_draws=["chain", "draw"]).values.T
            * self.precinct_pops
        )  # shape: num_samples x num_precincts

        # obtain samples of total votes summed across all precinct for the candidate for each group
        samples_of_votes_summed_across_district_gp1 = samples_converted_to_pops_gp1.sum(axis=1)
        samples_of_votes_summed_across_district_gp2 = samples_converted_to_pops_gp2.sum(axis=1)

        # obtain samples of the districtwide proportion of each demog. group voting for candidate
        self.sampled_voting_prefs[0] = (
            samples_of_votes_summed_across_district_gp1 / self.precinct_pops.sum()
        )  # sampled voted prefs across precincts
        self.sampled_voting_prefs[1] = (
            samples_of_votes_summed_across_district_gp2 / self.precinct_pops.sum()
        )  # sampled voted prefs across precincts

    def precinct_level_estimates(self):
        """If desired, we can return precinct-level estimates
        Returns:
            precinct_posterior_means: num_precincts x 2 (groups) x 2 (candidates)
            precinct_credible_intervals: num_precincts x 2 (groups) x 2 (candidates) x 2 (endpoints)
        """
        # TODO: make this output match r_by_c version in shape, num_precincts x 2 x 2
        percentiles = [2.5, 97.5]
        num_precincts = len(self.precinct_pops)

        # The stracking on the next line convers to a num_samples x num_precincts array
        precinct_level_samples_gp1 = (
            self.sim_trace["posterior"]["b_1"].stack(all_draws=["chain", "draw"]).values.T
        )
        precinct_posterior_means_gp1 = precinct_level_samples_gp1.mean(axis=0)
        precinct_credible_intervals_gp1 = np.percentile(
            precinct_level_samples_gp1, percentiles, axis=0
        ).T

        # The stracking on the next line convers to a num_samples x num_precincts array
        precinct_level_samples_gp2 = (
            self.sim_trace["posterior"]["b_2"].stack(all_draws=["chain", "draw"]).values.T
        )
        precinct_posterior_means_gp2 = precinct_level_samples_gp2.mean(axis=0)
        precinct_credible_intervals_gp2 = np.percentile(
            precinct_level_samples_gp2, percentiles, axis=0
        ).T  # num_precincts x 2

        precinct_posterior_means = np.empty((num_precincts, 2, 2))
        precinct_posterior_means[:, 0, 0] = precinct_posterior_means_gp1
        precinct_posterior_means[:, 0, 1] = 1 - precinct_posterior_means_gp1
        precinct_posterior_means[:, 1, 0] = precinct_posterior_means_gp2
        precinct_posterior_means[:, 1, 1] = 1 - precinct_posterior_means_gp2

        precinct_credible_intervals = np.empty((num_precincts, 2, 2, 2))
        precinct_credible_intervals[:, 0, 0, :] = precinct_credible_intervals_gp1
        precinct_credible_intervals[:, 0, 1, :] = 1 - precinct_credible_intervals_gp1
        precinct_credible_intervals[:, 1, 0, :] = precinct_credible_intervals_gp2
        precinct_credible_intervals[:, 1, 1, :] = 1 - precinct_credible_intervals_gp2

        return (precinct_posterior_means, precinct_credible_intervals)

    def plot_intervals_by_precinct(self):
        """Plot of point estimates and credible intervals for each precinct"""
        # TODO: Fix use of axes

        precinct_posterior_means, precinct_credible_intervals = self.precinct_level_estimates()
        precinct_posterior_means_gp1 = precinct_posterior_means[:, 0, 0]
        precinct_posterior_means_gp2 = precinct_posterior_means[:, 1, 0]
        precinct_credible_intervals_gp1 = precinct_credible_intervals[:, 0, 0, :]
        precinct_credible_intervals_gp2 = precinct_credible_intervals[:, 1, 0, :]

        plot_gp1 = plot_intervals_all_precincts(
            precinct_posterior_means_gp1,
            precinct_credible_intervals_gp1,
            self.candidate_name,
            self.precinct_names,
            self.group_names_for_display()[0],
        )
        plot_gp2 = plot_intervals_all_precincts(
            precinct_posterior_means_gp2,
            precinct_credible_intervals_gp2,
            self.candidate_name,
            self.precinct_names,
            self.group_names_for_display()[1],
        )

        return plot_gp1, plot_gp2

    def plot(self, axes=None):
        """kde, boxplot, and credible intervals
        Optional arguments:
        axes : list or tuple of matplotlib axis objects or None
            Default=None
            Length 2: (ax_box, ax_hist)
        """
        return plot_summary(
            self._voting_prefs_array(),
            self.group_names_for_display()[0],
            self.group_names_for_display()[1],
            self.candidate_name,
            axes=axes,
        )

    def precinct_level_plot(
        self,
        ax=None,
        alpha=1,
        show_all_precincts=False,
        precinct_names=None,
        plot_as_histograms=False,
    ):
        """Ridgeplots for precincts
        Optional arguments:
        ax                  :  matplotlib axes object
        show_all_precincts  :  If True, then it will show all ridge plots
                               (even if there are more than 50)
        alpha               : float
                                The opacity for the fill color in the
                                kdes / histograms
        precinct_names      :  Labels for each precinct (if not supplied, by
                               default we label each precinct with an integer
                               label, 1 to n)
        plot_as_histograms : bool, optional. Default is false. If true, plot
                                with histograms instead of kdes
        """
        voting_prefs_group1 = (
            self.sim_trace["posterior"]["b_1"].stack(all_draws=["chain", "draw"]).values.T
        )
        voting_prefs_group2 = (
            self.sim_trace["posterior"]["b_2"].stack(all_draws=["chain", "draw"]).values.T
        )
        group_names = self.group_names_for_display()
        if precinct_names is not None:
            precinct_idxs = np.arange(len(self.precinct_names))
            voting_prefs_group1 = voting_prefs_group1[:, precinct_idxs]
            voting_prefs_group2 = voting_prefs_group2[:, precinct_idxs]
        return plot_precincts(
            [voting_prefs_group1, voting_prefs_group2],
            group_names=group_names,
            candidate=self.candidate_name,
            alpha=alpha,
            precinct_labels=precinct_names,
            show_all_precincts=show_all_precincts,
            plot_as_histograms=plot_as_histograms,
            ax=ax,
        )



"""
Plotting functions for visualizing ei outputs

"""

import warnings
import seaborn as sns
import pandas as pd
from matplotlib import pyplot as plt
from matplotlib import ticker as mticker
from matplotlib.collections import PatchCollection
from matplotlib.patches import Rectangle
import matplotlib.patches as mpatches
import numpy as np
import scipy.stats as st

__all__ = [
    "plot_boxplots",
    "plot_conf_or_credible_interval",
    "plot_intervals_all_precincts",
    "plot_kdes",
    "plot_polarization_kde",
    "plot_precinct_scatterplot",
    "plot_precincts",
    "plot_summary",
    "tomography_plot",
]

PALETTE = "Dark2"  # set library-wide color palette
FONTSIZE = 20
TITLESIZE = 24
TICKSIZE = 15
FIGSIZE = (10, 5)
colors = sns.color_palette(PALETTE)


def size_ticks(ax, axis="x"):
    """
    Helper function to size the x- or ytick (numbers) of a matplotlib Axis

    Parameters
    ----------
    axis : string
        Either 'x' or 'y', specifies which axis's ticks are being sized
    ax: matplotlib axis object
    """
    if axis == "x":
        ax.set_xlim(0, 1)
        xticks = [round(xtick, 1) for xtick in ax.get_xticks()]
        ax.set_xticks(xticks)
        ax.set_xticklabels(xticks, size=TICKSIZE)
    elif axis == "y":
        ax.set_ylim(0, 1)
        yticks = [round(ytick, 1) for ytick in ax.get_yticks()]
        ax.set_yticks(yticks)
        ax.set_yticklabels(yticks, size=TICKSIZE)
    else:
        raise ValueError("You need to specify an 'x' or 'y' axis!")


def size_yticklabels(ax):
    """
    Helper function to size the ytick labels of a matplotlib Axis

    Parameters
    ----------
    ax : matplotlib axis object
    """
    ax.set_yticklabels(ax.get_yticklabels(), size=TICKSIZE)


def plot_single_ridgeplot(
    ax,
    group_prefs,
    colors,  # pylint: disable=redefined-outer-name
    alpha,
    z_init,
    trans,
    overlap=1.3,
    num_points=500,
):
    """Helper function for plot_precincts that plots a single ridgeplot (e.g.,
    for a single precinct for a given candidate.)

    Parameters
    ----------
    ax : matplotlib axis object
    group_prefs: [array]
        A list where each element is an array of estimates for support
        for the candidate from a group (array of floats, expected in [0,1])
    colors : array
        The (ordered) names of colors to use to fill ridgeplots
    z_init : float
        The initial value for the z-order (helps determine
        how plots get drawn over one another)
    trans
        The y-translation for this plot
    overlap : float, optional
        how much this ridgeplot may overlap with the ridgeplot
        above it
    num_points : int, optional
        The number of evenly spaced points in [0, 1] that we
        use to plot compute the KDE curve
    """
    x = np.linspace(0, 1, num_points)  # 500 points between 0 and 1 on the x-axis
    group_kdes = [st.gaussian_kde(group_pref) for group_pref in group_prefs]

    group_ys = []
    for i in range(len(group_prefs)):
        group_y = group_kdes[i](x)
        group_y = overlap * group_y / group_y.max()
        group_ys.append(group_y)

    for i, group_y in enumerate(group_ys):
        ax.fill_between(
            x,
            group_y + trans,
            trans,
            color=colors[i],
            alpha=alpha,
            zorder=z_init,
        )
        ax.plot(
            x,
            group_y + trans,
            color="black",
            linewidth=1,
            zorder=z_init + 1 + (2 * i),
        )


def plot_single_histogram(
    ax,
    group_prefs,
    colors,  # pylint: disable=redefined-outer-name
    alpha,
    z_init,
    trans,  # pylint: disable=redefined-outer-name
):
    """Helper function for plot_precincts that plots a single precinct histogram(s)
       (i.e.,for a single precinct for a given candidate.)

    Parameters
    ----------
    ax : matplotlib axis object
    group_prefs: [array]
        A list where each element is an array of estimates for support
        for the candidate from a group (array of floats, expected in [0,1])
    colors : array
        The (ordered) names of colors to use to fill ridgeplots
    z_init : float
        The initial value for the z-order (helps determine
        how plots get drawn over one another)
    trans
        The y-translation for this plot
    """

    bins = np.linspace(0, 1.0, num=20)
    for i, group_pref in enumerate(group_prefs):
        weights, bins = np.histogram(group_pref, bins=bins)
        weights = weights / weights.max()
        ax.hist(
            bins[:-1],
            bins=bins,
            weights=weights,
            bottom=trans,
            zorder=z_init + 1,
            color=colors[i],
            alpha=alpha,
            edgecolor="black",
        )


def plot_precincts(
    voting_prefs,
    group_names,
    candidate,
    alpha=1,
    precinct_labels=None,
    show_all_precincts=False,
    plot_as_histograms=False,
    ax=None,
):
    """Ridgeplots of sampled voting preferences for each precinct

    Parameters
    ----------
    voting_prefs : list of numpy arrays
        Each element has shape (# of samples x # of precincts) representing
        the samples of support for the given candidate among a given group
        in each precinct. Each element refers to a different group.
    group_names: list of str
        The demographic group names, for display in the legend
    candidate: str
        The candidate name
    alpha: float
        The opacity for the fill color in the kdes/histograms
    precinct_labels : list of str (optional)
        The names for each precinct
    show_all_precincts : bool, optional
        By default (show_all_precincts=False), we only show the first 50
        precincts. If show_all_precincts is True, we plot the ridgeplots
        for all precincts (i.e., one ridgeplot for every column in the
        voting_prefs matrices)
    plot_as_histograms : bool, optional
        Default: False If true, plot with histograms instead of kdes
    ax : Matplotlib axis object or None, optional
        Default=None

    Returns
    -------
    ax: Matplotlib axis object
    """
    N = voting_prefs[0].shape[1]
    if N > 50 and not show_all_precincts:
        warnings.warn(
            f"User attempted to plot {N} precinct-level voting preference "
            f"ridgeplots. Automatically restricting to first 50 precincts "
            f"(run with `show_all_precincts=True` to plot all precinct ridgeplots.)"
        )
        voting_prefs = [prefs[:, :50] for prefs in voting_prefs]
        if precinct_labels is not None:
            precinct_labels = precinct_labels[:50]
        N = 50
    if precinct_labels is None:
        precinct_labels = range(1, N + 1)

    legend_space = len(group_names) + 2
    if ax is None:
        # adapt height of plot to the number of precincts
        _, ax = plt.subplots(figsize=(FIGSIZE[0], 0.3 * (N + legend_space)))

    transposed_voting_prefs = [prefs.T for prefs in voting_prefs]
    iterator = zip(*transposed_voting_prefs)

    for idx, group_prefs in enumerate(iterator, 0):
        ax.plot([0], [idx])
        trans = ax.convert_yunits(idx)
        if plot_as_histograms:
            plot_single_histogram(ax, group_prefs, colors, alpha, 4 * (N - idx), trans)
        else:
            plot_single_ridgeplot(ax, group_prefs, colors, alpha, 4 * (N - idx), trans)
    for i in range(legend_space):
        # add `legend_space` number of lines to the top of the plot for legend
        ax.plot([0], [N + i])

    def replace_ticks_with_precinct_labels(value, pos):
        # pylint: disable=unused-argument
        # matplotlib axis tick formatter function
        idx = int(value)
        if idx < len(precinct_labels):
            return precinct_labels[idx]
        return value

    # replace y-axis ticks with precinct labels
    ax.set_yticks(np.arange(len(precinct_labels)))
    ax.yaxis.set_major_formatter(mticker.FuncFormatter(replace_ticks_with_precinct_labels))
    ax.set_title("Precinct level estimates of voting preferences", fontsize=TITLESIZE)
    ax.set_xlabel(f"Percent vote for {candidate}", fontsize=FONTSIZE)
    ax.set_ylabel("Precinct", fontsize=FONTSIZE)

    proxy_handles = [
        mpatches.Patch(color=colors[i], alpha=alpha, ec="black", label=group_names[i])
        for i in range(len(group_names))
    ]
    ax.legend(handles=proxy_handles, prop={"size": 14}, loc="upper center")
    ax.set_ylim(-1, ax.get_ylim()[1])
    size_ticks(ax, "x")
    return ax


def plot_boxplots(
    sampled_voting_prefs, group_names, candidate_names, plot_by="candidate", axes=None
):
    """
    Horizontal boxplots for r x c sets of samples between 0 and 1

    Parameters
    ----------
    sampled_voting_prefs : Numpy array
        Shape is: num_samples x r x c(where r=# of demographic groups,
        c=# of voting outcomes), representing the samples of the support
        for each candidate (or voting outcome) among each group (aggregated
        across all precincts)
    group_names : list of str
        Length = r (where r=# of demographic groups), the names of the
        demographic groups (order should match order of the columns in
        sampled_voting_prefs)
    candidate_names : list of str
        Length = c (where c=# of voting outcomes), the names of the candidates
        or voting outcomes (order should match order of the last dimension of
        sampled_voting_prefs)
    plot_by : {"candidate", "group"}
        (Default='candidate')
        If 'candidate', make one plot per candidate, with each plot showing the boxplots of
        estimates of voting preferences of all groups. If 'group', one plot
        per group, with each plot showing the boxplots of estimates of voting
        preferences for all candidates.
    axes : list of Matplotlib axis object or None
        Default=None.
        If not None and plot_by is 'candidate', should have length c (number of candidates).
        If plot_by is 'group', should have length r (number of groups)

    Returns
    -------
    ax : Matplotlib axis object
        Has c subplots, each showing the sampled voting preference of each of r groups.

    Notes
    -----
    If passing existing axes within a subplot, consider using, e.g.,
    plt.subplots_adjust(hspace=0.75) to make control spacing
    """

    _, num_groups, num_candidates = sampled_voting_prefs.shape
    if plot_by == "candidate":
        num_plots = num_candidates
        num_boxes_per_plot = num_groups
        titles = candidate_names
        legend = group_names
        support = "for"
        if axes is None:
            _, axes = plt.subplots(num_candidates, figsize=FIGSIZE)

    elif plot_by == "group":
        num_plots = num_groups
        num_boxes_per_plot = num_candidates
        titles = group_names
        sampled_voting_prefs = np.swapaxes(sampled_voting_prefs, 1, 2)
        legend = candidate_names
        support = "among"
        if axes is None:
            _, axes = plt.subplots(num_groups, figsize=FIGSIZE)
    else:
        raise ValueError("plot_by must be 'group' or 'candidate' (default: 'candidate')")
    plt.gcf().subplots_adjust(hspace=1)

    for plot_idx in range(num_plots):
        samples_df = pd.DataFrame(
            {legend[i]: sampled_voting_prefs[:, i, plot_idx] for i in range(num_boxes_per_plot)}
        )
        if num_plots > 1:
            ax = axes[plot_idx]
        else:
            ax = axes
        sns.boxplot(data=samples_df, orient="h", whis=[2.5, 97.5], ax=ax, palette=colors)
        ax.set_title(f"Support {support} {titles[plot_idx]}", fontsize=TITLESIZE)
        ax.tick_params(axis="y", left=False)  # remove y axis ticks
        size_ticks(ax, "x")
        size_yticklabels(ax)

    return ax


def plot_summary(
    sampled_voting_prefs,
    group1_name,
    group2_name,
    candidate_name,
    axes=None,
):
    """Plot KDE, histogram, and boxplot for 2x2 case

    Parameters
    ----------
    sampled_voting_prefs : array
        num_samples x 2 x 1 Samples of estimated voting preferences (support)
        of group 1 for the candidate.
    group1_name : str
        Name of group 1 (used for legend of plot)
    group2_name : str
        Name of group 2 (used for legend of plot)
    candidate_name: str
        The name of the candidate
    axes : list or tuple of matplotlib axis objects or None
        Default=None
        Length 2: (ax_box, ax_hist)

    Returns
    -------
    ax_box : Matplotlib axis object
    ax_hist : Matplotlib axis object
    """
    if axes is None:
        _, (ax_hist, ax_box) = plt.subplots(
            2,
            sharex=True,
            figsize=FIGSIZE,
            gridspec_kw={"height_ratios": (0.85, 0.15)},
        )
    else:
        ax_hist, ax_box = axes
    size_ticks(ax_box, "x")
    ax_box.set_title("")
    ax_box.set_xlabel(f"Support for {candidate_name}", fontsize=FONTSIZE)
    sns.despine(ax=ax_hist)
    sns.despine(ax=ax_box, left=True)
    # plot custom boxplot, with two boxplots in the same row
    plot_props = {"fliersize": 5, "linewidth": 2, "whis": [2.5, 97.5]}
    flier1_props = {"marker": "o", "markerfacecolor": colors[0], "alpha": 0.5}
    flier2_props = {"marker": "d", "markerfacecolor": colors[1], "alpha": 0.5}
    sns.boxplot(
        x=sampled_voting_prefs[:, 0, 0],
        orient="h",
        color=colors[0],
        ax=ax_box,
        flierprops=flier1_props,
        **plot_props,
    )
    sns.boxplot(
        x=sampled_voting_prefs[:, 1, 0],
        orient="h",
        color=colors[1],
        ax=ax_box,
        flierprops=flier2_props,
        **plot_props,
    )
    ax_box.tick_params(axis="y", left=False)  # remove y axis ticks

    # plot distribution
    plot_kdes(sampled_voting_prefs, [group1_name, group2_name], [candidate_name], axes=ax_hist)
    ax_hist.set_title("EI Summary", fontsize=TITLESIZE)
    plt.subplots_adjust(hspace=0.005)
    return (ax_box, ax_hist)


def plot_precinct_scatterplot(ei_runs, run_names, candidate, demographic_group="all", ax=None):
    """
    Given two RxC EI runs, plot precinct-by-precinct comparison of preferences
    for a given candidate from a given demographic group.

    Parameters
    ----------
    ei_runs: array
        Length = 2
        Element Type = RowByColumnEI
    run_names: array
        Length = 2
        Element Type = string
        Name of each EI run (in the same order as ei_runs!)
    candidate: string
        Must be a candidate common to both EI runs
    demographic_group: string
        Must be a demographic group common to both EI runs, or "all",
        which plots and labels each demographic group onto the same axes.

    Returns
    -------
    ax: Matplotlib axis object
    """
    if ax is None:
        _, ax = plt.subplots(1, figsize=(2 * FIGSIZE[0], 2 * FIGSIZE[1]))
    ax.set_xlim((0, 1))
    ax.set_ylim((0, 1))

    prec_means1, _ = ei_runs[0].precinct_level_estimates()
    prec_means2, _ = ei_runs[1].precinct_level_estimates()

    # Set group names and candidates in case runs are TwoByTwoEI
    if not hasattr(ei_runs[0], "demographic_group_names"):  # then is TwoByTwoEI
        demographic_group_names1 = list(ei_runs[0].group_names_for_display())
        candidate_names1 = [
            ei_runs[0].candidate_name,
            "not " + ei_runs[0].candidate_name,
        ]
    else:
        demographic_group_names1 = ei_runs[0].demographic_group_names
        candidate_names1 = ei_runs[0].candidate_names
    if not hasattr(ei_runs[1], "demographic_group_names"):  # then it is TwoByTwoEI
        demographic_group_names2 = list(ei_runs[1].group_names_for_display())
        candidate_names2 = [
            ei_runs[1].candidate_name,
            "not " + ei_runs[1].candidate_name,
        ]
    else:
        demographic_group_names2 = ei_runs[1].demographic_group_names
        candidate_names2 = ei_runs[1].candidate_names

    common_groups = [g for g in demographic_group_names1 if g in demographic_group_names2]

    group_dict = {}
    for group in common_groups:
        group_dict[group] = (
            prec_means1[
                :,
                demographic_group_names1.index(group),
                candidate_names1.index(candidate),
            ],
            prec_means2[
                :,
                demographic_group_names2.index(group),
                candidate_names2.index(candidate),
            ],
        )

    if demographic_group == "all":
        for k, (x_vals, y_vals) in group_dict.items():
            sns.scatterplot(x=x_vals, y=y_vals, label=k)
    else:
        sns.scatterplot(
            x=group_dict[demographic_group][0],
            y=group_dict[demographic_group][1],
            label=demographic_group,
        )
    sns.lineplot(x=[0, 1], y=[0, 1], alpha=0.5, color="grey")
    ax.set_title(
        f"{run_names[0]} vs. {run_names[1]}\n Predicted support for {candidate}",
        fontsize=TITLESIZE,
    )
    ax.set_xlabel(f"Support for {candidate} (from {run_names[0]})", fontsize=FONTSIZE)
    ax.set_ylabel(f"Support for {candidate} (from {run_names[1]})", fontsize=FONTSIZE)
    ax.legend()
    size_ticks(ax, "x")
    size_ticks(ax, "y")

    return ax


def plot_margin_kde(group, candidates, samples, thresholds, percentile, show_threshold, ax):
    """
    Plots a kde for the margin between two candidates among a given demographic group

    Parameters:
    -----------
    samples: array
        samples of the differences in voting preferences (candidate 1 - candidate 2)
    thresholds: array
        a list of thresholds for the difference in voting patterns between two groups
    group: str
        the name of the demographic group in question
    candidates : list of str
        the names of the two candidates in question
    show_threshold: bool
        if true, add vertical lines at the threshold on the plot

    Returns
    -------
    ax: Matplotlib axis object
    """
    if ax is None:
        _, ax = plt.subplots(figsize=FIGSIZE)

    sns.histplot(
        samples,
        kde=True,
        ax=ax,
        element="step",
        stat="density",
        color="steelblue",
        linewidth=0,
    )
    ax.set_ylabel("Density", fontsize=FONTSIZE)
    if len(thresholds) == 1:
        threshold_string = f"> {thresholds[0]:.2f}"
    else:
        threshold_string = f"in [{thresholds[0]:.2f}, {thresholds[1]:.2f}]"
    if show_threshold:
        for threshold in thresholds:
            ax.axvline(threshold, c="gray")
        if len(thresholds) == 2:
            ax.axvspan(thresholds[0], thresholds[1], facecolor="gray", alpha=0.2)
        else:
            ax.axvspan(thresholds[0], 1, facecolor="gray", alpha=0.2)
        ax.text(
            thresholds[-1] + 0.05,
            0.5,
            f"Prob (margin {threshold_string} ) = {percentile:.1f}%",
            fontsize=FONTSIZE,
        )

    ax.set_title(f"{candidates[0]} - {candidates[1]} margin among {group}", fontsize=TITLESIZE)
    ax.set_xlabel(f"{group} support for {candidates[0]} - {candidates[1]}", fontsize=FONTSIZE)
    ax.set_xlim((-1, 1))
    xticks = ax.get_xticks()
    ax.set_xticks(xticks)
    ax.set_xticklabels(xticks, size=TICKSIZE)


def plot_polarization_kde(
    diff_samples,
    thresholds,
    probability,
    groups,
    candidate_name,
    show_threshold=False,
    ax=None,
    color="steelblue",
):
    """
    Plots a kde for the differences in voting preferences between two groups

    Parameters:
    -----------
    diff_samples: array
        samples of the differences in voting preferences (group_complement - group)
    probability: float
        the probability that (group_complement - group) > threshold
    thresholds: array
        a list of thresholds for the difference in voting patterns between two groups
    groups: list
        the names of the two groups being compared
    candidate_name : string
        the name of the candidate or voting outcome whose support is shown in the kde
    show_threshold: bool
        if true, add a vertical line at the threshold on the plot and display the associated
        tail probability
    color: str
        specifies a color for matplotlib to be used in the histogram/kde

    Returns
    -------
    ax: Matplotlib axis object
    """

    if ax is None:
        _, ax = plt.subplots(figsize=FIGSIZE)

    sns.histplot(
        diff_samples,
        kde=True,
        ax=ax,
        element="step",
        stat="density",
        label=groups[0] + " - " + groups[1],
        color=color,
        linewidth=0,
    )
    ax.set_ylabel("Density", fontsize=FONTSIZE)
    if len(thresholds) == 1:
        threshold_string = f"> {thresholds[0]:.2f}"
    else:
        threshold_string = f"in [{thresholds[0]:.2f}, {thresholds[1]:.2f}]"
    if show_threshold:
        for threshold in thresholds:
            ax.axvline(threshold, c="gray")
        if len(thresholds) == 2:
            ax.axvspan(thresholds[0], thresholds[1], facecolor="gray", alpha=0.2)
        else:
            ax.axvspan(thresholds[0], 1, facecolor="gray", alpha=0.2)
        ax.text(
            thresholds[-1] + 0.05,
            0.5,
            f"Prob (difference {threshold_string} ) = {probability:.1f}%",
            fontsize=FONTSIZE,
        )

    ax.set_title(f"Polarization KDE for {candidate_name}", fontsize=TITLESIZE)
    ax.set_xlabel(f"({groups[0]} - {groups[1]}) support for {candidate_name}", fontsize=FONTSIZE)
    ax.set_xlim((-1, 1))
    xticks = ax.get_xticks()
    ax.set_xticks(xticks)
    ax.set_xticklabels(xticks, size=TICKSIZE)

    return ax


def plot_kdes(sampled_voting_prefs, group_names, candidate_names, plot_by="candidate", axes=None):
    """
    Plot a kernel density plot for prefs of voting groups for each candidate

    Parameters
    ----------
    sampled_voting_prefs : numpy array
        Shape: num_samples x r x c (where r = # of demographic groups, c= #
        of candidates or voting outcomes). Gives samples of support from each group
        for each candidate. NOTE: for a 2 x 2 case where we have just two candidates/outcomes
        and want only one plot, have c=1
    group_names : list of str
        Names of the demographic groups (length r)
    candidate_names : list of str
        Names of the candidates or voting outcomes (length c)
    plot_by : {"candidate", "group"}
        (Default='candidate')
        If 'candidate', make one plot per candidate, with each plot showing the kernel
        density estimates of voting preferences of all groups. If 'group', one plot
        per group, with each plot showing the kernel density estimates of voting
        preferences for all candidates.
    axes : list of Matplotlib axis object or None
        Default=None.
        If not None and plot_by is 'candidate', should have length c (number of candidates).
        If plot_by is 'group', should have length r (number of groups)

    Returns
    -------
    ax : Matplotlib axis object
    """

    _, num_groups, num_candidates = sampled_voting_prefs.shape
    if plot_by == "candidate":
        num_plots = num_candidates
        num_kdes_per_plot = num_groups
        titles = candidate_names
        legend = group_names
        support = "for"
        if axes is None:
            _, axes = plt.subplots(num_candidates, figsize=FIGSIZE, sharex=True)
            plt.subplots_adjust(hspace=0.5)
    elif plot_by == "group":
        num_plots = num_groups
        num_kdes_per_plot = num_candidates
        titles = group_names
        sampled_voting_prefs = np.swapaxes(sampled_voting_prefs, 1, 2)
        legend = candidate_names
        support = "among"
        if axes is None:
            _, axes = plt.subplots(num_groups, figsize=FIGSIZE, sharex=True)
            plt.subplots_adjust(hspace=0.5)
    else:
        raise ValueError("plot_by must be 'group' or 'candidate' (default: 'candidate')")

    middle_plot = int(np.floor(num_plots / 2))
    for plot_idx in range(num_plots):
        if num_plots > 1:
            ax = axes[plot_idx]
            axes[middle_plot].set_ylabel("Probability Density", fontsize=FONTSIZE)
        else:
            ax = axes
            axes.set_ylabel("Probability Density", fontsize=FONTSIZE)
        ax.set_title(f"Support {support} " + titles[plot_idx], fontsize=TITLESIZE)
        ax.set_xlim((0, 1))
        size_ticks(ax, "x")

        for kde_idx in range(num_kdes_per_plot):
            sns.histplot(
                sampled_voting_prefs[:, kde_idx, plot_idx],
                kde=True,
                ax=ax,
                stat="density",
                element="step",
                label=legend[kde_idx],
                color=colors[kde_idx],
                linewidth=0,
            )
            ax.set_ylabel("")

    if num_plots > 1:
        axes[middle_plot].legend(bbox_to_anchor=(1, 1), loc="upper left", prop={"size": 12})
    else:
        ax.legend(prop={"size": 12})
    return axes


def plot_conf_or_credible_interval(intervals, group_names, candidate_name, title, ax=None):
    """
    Plot confidence of credible interval for two different groups

    Parameters
    ----------
    intervals : list of arrays or tuple
        Length is number of demographic groups, each element in the array
        is a length-two array that gives
        (lower, upper) bounds for credible or confidence interval
        for support from group 1 for the candidate of interest
    groups_names : list of str
        Names of groups (ordered to match order of intervals), for plot legend
    candidate_name : str
        Name of candidate (or voting outcome) whose support is to be plotted
    title : str
        Title of plot
    ax : Matplotlib axis object or None, optional
        Default=None

    Returns
    -------
    ax : Matplotlib axis object
    """

    if ax is None:
        _, ax = plt.subplots(figsize=FIGSIZE)

    int_heights = 0.2 * np.arange(len(group_names), 0, -1)
    ax.set_ylim(0, (len(group_names) + 1) * 0.2)
    ax.set_title(title, fontsize=TITLESIZE)
    ax.set_xlabel(f"Support for {candidate_name}", fontsize=FONTSIZE)
    ax.set(
        xlim=(0, 1),
        frame_on=True,
        aspect=0.3,
    )

    ax.get_xaxis().tick_bottom()
    ax.axes.get_yaxis().set_visible(False)
    ax.grid()
    for idx, group_name in enumerate(group_names):
        ax.text(1, int_heights[idx], f" {group_name}", fontsize=FONTSIZE)
        ax.plot(
            intervals[idx],
            [int_heights[idx], int_heights[idx]],
            linewidth=10,
            alpha=0.8,
            color=colors[idx],
        )
    size_ticks(ax, "x")

    return ax


def plot_intervals_all_precincts(
    point_estimates,
    intervals,
    candidate_name,
    precinct_labels,
    title,
    ax=None,
    show_all_precincts=False,
):
    """
    Plot intervals&point estimates of support for candidate, sorted by point estimates for precincts

    Parameters
    ----------
    point_estimates: array
        Array of length num_precincts, each element is the estimate of support for the
        candidate among the demographic group of interest in that precinct
    intervals: array
        Array of arrays (shape num_precincts, each elt an array of length 2)
        Each element is an array of length two giving the lower, upper limits of a credible
        or confidence interval for support for the candidate in the precinct
    candidate_name: str
        Name of candidate, used in x axis label
    title: str
        Title for plot
    ax : Matplotlib axis object or None, optional
        Default=None
    show_all_precincts: bool, optional
        (default=False). If True, show estimates&intervals for all precincts. If False,
        show only the first 50, for readibility.

    Returns
    -------
    ax : Matplotlib axis object
    """
    num_intervals = len(point_estimates)

    if num_intervals > 50 and not show_all_precincts:
        warnings.warn(
            f"User attempted to plot {num_intervals} precinct-level voting preference "
            f"ridgeplots. Automatically restricting to first 50 precincts "
            f"(run with `show_all_precincts=True` to plot all precinct ridgeplots.)"
        )
        point_estimates = point_estimates[:50]
        intervals = intervals[:, :50]
        if precinct_labels is not None:
            precinct_labels = precinct_labels[:50]
        num_intervals = 50

    int_heights = 20 * np.arange(num_intervals) + 20
    tot_height = int_heights[-1] + 20
    int_heights = tot_height - int_heights

    if ax is None:
        _, ax = plt.subplots(
            1, frameon=False, constrained_layout=True, figsize=(16, num_intervals / 4)
        )

    if precinct_labels is None:
        precinct_labels = range(num_intervals)

    ax.set(
        title=title,
        xlim=(0, 1),
        ylim=(0, tot_height),
        xlabel=f"Support for {candidate_name}",
        frame_on=False,
    )

    ax.get_xaxis().tick_bottom()
    ax.axes.get_yaxis().set_visible(False)

    point_estimates, intervals, precinct_labels = zip(
        *sorted(zip(point_estimates, intervals, precinct_labels))
    )
    bars = []

    for point_estimate, interval, precinct_label, int_height in zip(
        point_estimates, intervals, precinct_labels, int_heights
    ):
        width = interval[1] - interval[0]
        height = 16
        lower_left = (interval[0], int_height)
        bars.append(Rectangle(lower_left, width, height))

        ax.scatter(point_estimate, int_height + height / 2, s=30, alpha=1, c="k")
        ax.text(1, int_height + height / 2, precinct_label, fontsize=12)

    pc_bars = PatchCollection(bars, facecolor="gray", alpha=0.6)
    ax.add_collection(pc_bars)

    return ax


def tomography_plot(
    group_fraction,
    votes_fraction,
    demographic_group_name,
    candidate_name,
    ax=None,
    c="b",
    **plot_kwargs,
):
    """Tomography plot (basic), applicable for 2x2 ei

    Parameters
    ----------
    group_fraction : array
        Array of length num_precincts, giving fraction of population of interest
        in each precinct represented by demographic group of interest
    votes_fraction: array
        An array of length num_precincts giving fraction  of votes in each
        precinct for candidate of interest
    demographic_group_name : str
        Name of demographic group of interest
    candidate_name : str
        Name of candidate or voting outcome of interest
    ax : Matplotlib axis object or None, optional
        Default=None
    c : specifies a color for Matplotlib, optional
        Default="b"
    **plot_kwargs
        Additional keyword arguments to be passed to matplotlib.Axes.plot()

    Returns
    -------
    ax : Matplotlib axis object
    """

    if ax is None:
        _, ax = plt.subplots()

    num_precincts = len(group_fraction)
    b_1 = np.linspace(0, 1, 200)
    ax.set_xlim((0, 1))
    ax.set_ylim((0, 1))
    ax.set_aspect("equal", adjustable="box")
    ax.set_xlabel(f"voter pref of {demographic_group_name} for {candidate_name}")
    ax.set_ylabel(f"voter pref of non-{demographic_group_name} for {candidate_name}")
    for i in range(num_precincts):
        b_2 = (votes_fraction[i] - b_1 * group_fraction[i]) / (1 - group_fraction[i])
        ax.plot(b_1, b_2, c=c, **plot_kwargs)
    return ax



"""test r-by-c specific plotting"""

import pytest
import numpy as np
from pyei import data
from pyei.plot_utils import plot_precinct_scatterplot
from pyei.r_by_c import RowByColumnEI

# from pyei.io_utils import to_netcdf, from_netcdf


@pytest.fixture(scope="session")
def example_r_by_c_data():
    """trimmed santa clara dataset"""
    sc_data = data.Datasets.Santa_Clara.to_dataframe()
    sc_data = sc_data.iloc[:10, :]
    precinct_pops = np.array(sc_data["total2"])
    votes_fractions = np.array(sc_data[["pct_for_hardy2", "pct_for_kolstad2", "pct_for_nadeem2"]]).T
    candidate_names = ["Hardy", "Kolstad", "Nadeem"]
    group_fractions = np.array(
        sc_data[["pct_ind_vote", "pct_e_asian_vote", "pct_non_asian_vote"]]
    ).T
    demographic_group_names = ["ind", "e_asian", "non_asian"]
    return {
        "group_fractions": group_fractions,
        "votes_fractions": votes_fractions,
        "precinct_pops": precinct_pops,
        "demographic_group_names": demographic_group_names,
        "candidate_names": candidate_names,
    }


@pytest.fixture(scope="session")
def example_r_by_c_data_asym():
    """trimmed santa clara dataset with r not equal to c"""
    sc_data = data.Datasets.Santa_Clara.to_dataframe()
    sc_data = sc_data.iloc[:10, :]
    precinct_pops = np.array(sc_data["total2"])
    votes_fractions = np.array(sc_data[["pct_for_hardy2", "pct_for_kolstad2", "pct_for_nadeem2"]]).T
    candidate_names = ["Hardy", "Kolstad", "Nadeem"]
    group_fractions = np.array(sc_data[["pct_asian_vote", "pct_non_asian_vote"]]).T
    demographic_group_names = ["asian", "non_asian"]
    return {
        "group_fractions": group_fractions,
        "votes_fractions": votes_fractions,
        "precinct_pops": precinct_pops,
        "demographic_group_names": demographic_group_names,
        "candidate_names": candidate_names,
    }


def example_r_by_c_ei(example_r_by_c_data, model_name):  # pylint: disable=redefined-outer-name
    """Run this to generate an EI instance"""

    ei_ex = RowByColumnEI(model_name=model_name)
    ei_ex.fit(
        example_r_by_c_data["group_fractions"],
        example_r_by_c_data["votes_fractions"],
        example_r_by_c_data["precinct_pops"],
        example_r_by_c_data["demographic_group_names"],
        example_r_by_c_data["candidate_names"],
        draws=100,
        tune=100,
    )
    return ei_ex


@pytest.fixture(scope="session")
def two_r_by_c_ei_runs(
    example_r_by_c_data, example_r_by_c_data_asym
):  # pylint: disable=redefined-outer-name
    """use the EI Factory to fix two EI instances for our tests"""
    example_ei_r_by_c_1 = example_r_by_c_ei(example_r_by_c_data, "multinomial-dirichlet")
    example_ei_r_by_c_2 = example_r_by_c_ei(example_r_by_c_data, "multinomial-dirichlet-modified")
    example_ei_r_by_c_asym = example_r_by_c_ei(example_r_by_c_data_asym, "multinomial-dirichlet")
    return [example_ei_r_by_c_1, example_ei_r_by_c_2, example_ei_r_by_c_asym]


def test_ei_r_by_c_summary(two_r_by_c_ei_runs):  # pylint: disable=redefined-outer-name
    example_r_by_c_ei = two_r_by_c_ei_runs[0]  # pylint: disable=redefined-outer-name
    assert isinstance(example_r_by_c_ei.summary(), str)


# @TODO: this test fails on github but not locally - fix
# def test_io_utils(two_r_by_c_ei_runs):  # pylint: disable=redefined-outer-name
#     example_r_by_c_ei = two_r_by_c_ei_runs[0]  # pylint: disable=redefined-outer-name
#     model_name_orig = example_r_by_c_ei.model_name
#     to_netcdf(example_r_by_c_ei, "example.nc")
#     reloaded_ei = from_netcdf("example.nc")
#     assert isinstance(reloaded_ei.summary(), str)  # check that summary string is there
#     assert model_name_orig == reloaded_ei.model_name  # check that model data came with


def test_ei_calculate_turnout_adjusted_samples(
    two_r_by_c_ei_runs,
):  # pylint: disable=redefined-outer-name
    def calculate_turnout_adjust_samples_basic(
        non_adjusted_samples, abstain_column_name, candidate_names
    ):
        # non_adjusted_samples = self.sim_trace.get_values("b") num_samples x num_precincts x r x c
        num_samples, num_precincts, num_rows, _ = non_adjusted_samples.shape

        abstain_column_index = candidate_names.index(abstain_column_name)

        turnout_adjusted_samples = np.delete(non_adjusted_samples, abstain_column_index, axis=3)

        for r_idx in range(num_rows):
            for samp_idx in range(num_samples):
                for p_idx in range(num_precincts):
                    # for c_idx in range(c-1):
                    turnout_adjusted_samples[samp_idx, p_idx, r_idx, :] = (
                        turnout_adjusted_samples[samp_idx, p_idx, r_idx, :]
                        / turnout_adjusted_samples[samp_idx, p_idx, r_idx, :].sum()
                    )

        return turnout_adjusted_samples

    example_r_by_c_ei = two_r_by_c_ei_runs[0]  # pylint: disable=redefined-outer-name
    non_adjusted_samples = np.transpose(
        example_r_by_c_ei.sim_trace["posterior"]["b"].stack(all_draws=["chain", "draw"]).values,
        axes=(3, 0, 1, 2),
    )
    test_adj_samples = calculate_turnout_adjust_samples_basic(
        non_adjusted_samples, "Hardy", example_r_by_c_ei.candidate_names
    )

    example_r_by_c_ei.calculate_turnout_adjusted_summary(["Hardy"])
    turnout_adjusted_samps_pyei = example_r_by_c_ei.turnout_adjusted_samples
    assert np.all(np.isclose(test_adj_samples, turnout_adjusted_samps_pyei))


def test_computation_of_districtwide_samples(
    two_r_by_c_ei_runs,
):  # pylint: disable=redefined-outer-name:  # pylint: disable=redefined-outer-name:
    """
    Testing calculations of RowByColumnEI.sampled_voting_prefs
    """
    ei_ex = two_r_by_c_ei_runs[0]

    def calculate_districtwide_samples_basic(samples, demographic_group_fractions, precinct_pops):
        # group fraction has shape: r x num_precincts
        demographic_group_counts = np.transpose(
            demographic_group_fractions * precinct_pops
        )  # num_precincts x r
        num_samples, _, num_rows, num_cols = samples.shape
        districtwide_prefs = np.empty((num_samples, num_rows, num_cols))
        for r_idx in range(num_rows):
            for samp_idx in range(num_samples):
                for c_idx in range(num_cols):
                    districtwide_prefs[samp_idx, r_idx, c_idx] = (
                        samples[samp_idx, :, r_idx, c_idx] * demographic_group_counts[:, r_idx]
                    ).sum() / (demographic_group_counts[:, r_idx]).sum()
        return districtwide_prefs

    test_districtwide_prefs = calculate_districtwide_samples_basic(
        np.transpose(
            ei_ex.sim_trace["posterior"]["b"].stack(all_draws=["chain", "draw"]).values,
            axes=(3, 0, 1, 2),
        ),
        ei_ex.demographic_group_fractions,
        ei_ex.precinct_pops,
    )
    np.all(np.isclose(test_districtwide_prefs, ei_ex.sampled_voting_prefs))


def test_candidate_of_choice_report(
    two_r_by_c_ei_runs,
):  # pylint: disable=redefined-outer-name
    example_r_by_c_ei = two_r_by_c_ei_runs[0]  # pylint: disable=redefined-outer-name
    candidate_preference_rate_dict = example_r_by_c_ei.candidate_of_choice_report(
        verbose=True, non_candidate_names=None
    )
    assert 0 <= candidate_preference_rate_dict["e_asian", "Kolstad"] < 1


def test_candidate_of_choice_polarization_report(
    two_r_by_c_ei_runs,
):  # pylint: disable=redefined-outer-name)
    example_r_by_c_ei = two_r_by_c_ei_runs[0]  # pylint: disable=redefined-outer-name
    candidate_differ_rate_dict = example_r_by_c_ei.candidate_of_choice_polarization_report(
        verbose=True, non_candidate_names=None
    )
    assert candidate_differ_rate_dict["non_asian", "e_asian"] > 0.4


def test_polarization_report(
    two_r_by_c_ei_runs,
):  # pylint: disable=redefined-outer-name
    example_r_by_c_ei = two_r_by_c_ei_runs[0]  # pylint: disable=redefined-outer-name
    groups = ["e_asian", "non_asian"]
    candidate = "Kolstad"
    prob_20 = example_r_by_c_ei.polarization_report(groups, candidate, threshold=0.2)
    prob_40 = example_r_by_c_ei.polarization_report(groups, candidate, threshold=0.4)
    thresh_95_range = example_r_by_c_ei.polarization_report(groups, candidate, percentile=95)
    thresh_90_range = example_r_by_c_ei.polarization_report(groups, candidate, percentile=90)

    assert prob_20 >= prob_40
    assert thresh_95_range[1] > thresh_95_range[0]
    assert thresh_95_range[1] - thresh_95_range[0] >= thresh_90_range[1] - thresh_90_range[0]


# TEST PLOTTING


def test_ei_r_by_c_precinct_scatterplot(
    two_r_by_c_ei_runs,
):  # pylint: disable=redefined-outer-name
    all_demographics_ax = plot_precinct_scatterplot(
        two_r_by_c_ei_runs, ["Run 1", "Run 2"], "Kolstad"
    )
    just_ind_ax = plot_precinct_scatterplot(two_r_by_c_ei_runs, ["Run 1", "Run 2"], "Nadeem", "ind")
    assert all_demographics_ax is not None
    assert just_ind_ax is not None


def test_ei_r_by_c_boxplots(two_r_by_c_ei_runs):  # pylint: disable=redefined-outer-name
    # TODO: maybe uncouple this to test the plot utils piece alone
    example_r_by_c_ei = two_r_by_c_ei_runs[0]  # pylint: disable=redefined-outer-name
    assert example_r_by_c_ei.plot_boxplots() is not None
    assert example_r_by_c_ei.plot_boxplots(plot_by="group") is not None
    with pytest.raises(ValueError):
        example_r_by_c_ei.plot_boxplots(plot_by="grupo")


def test_ei_r_by_c_kdes(two_r_by_c_ei_runs):  # pylint: disable=redefined-outer-name
    # TODO: maybe uncouple this to test the plot utils piece alone
    example_r_by_c_ei = two_r_by_c_ei_runs[0]  # pylint: disable=redefined-outer-name
    assert example_r_by_c_ei.plot_kdes(plot_by="candidate") is not None
    assert example_r_by_c_ei.plot_kdes(plot_by="group") is not None
    with pytest.raises(ValueError):
        example_r_by_c_ei.plot_kdes(plot_by="grupo")


def test_ei_r_by_c_intervals_by_precinct(
    two_r_by_c_ei_runs,
):  # pylint: disable=redefined-outer-name
    example_r_by_c_ei = two_r_by_c_ei_runs[0]
    assert example_r_by_c_ei.plot_intervals_by_precinct("e_asian", "Kolstad") is not None
    with pytest.raises(ValueError):
        example_r_by_c_ei.plot_intervals_by_precinct("e_asian", "Kolstaad")
        example_r_by_c_ei.plot_intervals_by_precinct("ibnd", "Hardy")


def test_plot_polarization_kde(
    two_r_by_c_ei_runs,
):  # pylint: disable=redefined-outer-name
    example_r_by_c_ei = two_r_by_c_ei_runs[0]  # pylint: disable=redefined-outer-name
    groups = ["ind", "e_asian"]
    candidate = "Kolstad"
    percentile_ax = example_r_by_c_ei.plot_polarization_kde(
        groups, candidate, threshold=0.4, show_threshold=True
    )
    percentile_ax_2 = example_r_by_c_ei.plot_polarization_kde(
        groups, candidate, threshold=0.4, show_threshold=True
    )
    threshold_ax = example_r_by_c_ei.plot_polarization_kde(
        groups, candidate, percentile=95, show_threshold=True
    )
    assert percentile_ax is not None
    print("done 1")
    assert percentile_ax_2 is not None
    print("done 2")
    assert threshold_ax is not None
    print("done 3")
    with pytest.raises(ValueError):
        example_r_by_c_ei.plot_polarization_kde(
            groups, candidate, threshold=0.4, percentile=95, show_threshold=True
        )
    print("done")


def test_plot_margin_kde(two_r_by_c_ei_runs):  # pylint: disable=redefined-outer-name
    example_r_by_c_ei = two_r_by_c_ei_runs[0]  # pylint: disable=redefined-outer-name
    example_r_by_c_ei.plot_margin_kde("ind", ["Hardy", "Nadeem"], threshold=0.1)



"""
Models and fitting for rxc methods
where r and c are greater than or
equal to 2

TODO: Investigate better or reparametrized priors for multinomial-dir
TODO: Greiner-Quinn Model
TODO: Refactor to integrate with two_by_two
"""

import warnings
import pymc as pm
import numpy as np
from .plot_utils import (
    plot_boxplots,
    plot_kdes,
    plot_intervals_all_precincts,
    plot_polarization_kde,
    plot_margin_kde,
    plot_precincts,
)
from .r_by_c_models import ei_multinom_dirichlet, ei_multinom_dirichlet_modified
from .r_by_c_utils import check_dimensions_of_input
from .greiner_quinn_gibbs_sampling import pyei_greiner_quinn_sample

__all__ = ["RowByColumnEI"]


class RowByColumnEI:  # pylint: disable=too-many-instance-attributes
    """
    Fitting and plotting for RxC models, fit via sampling
    """

    def __init__(self, model_name, **additional_model_params):
        """
        Parameters:
        -----------
        model_name: str
            The name of the model to use. Currently supported: multinomial-dirichlet,
            multinomial-dirichlet-modified
        additional_model_params: optional
            Settings for model hyperparameters, if desired to change default. See
            model function documentation for hyperparameter options
        """
        # TODO: implement greiner quinn
        # TODO: model_name as enumeration
        # TODO: clean up instance variables

        self.model_name = model_name
        self.additional_model_params = additional_model_params

        self.demographic_group_fractions = None
        self.votes_fractions = None
        self.precinct_pops = None
        self.precinct_names = None
        self.demographic_group_names = None
        self.candidate_names = None
        self.sim_model = None
        self.sim_trace = None
        self.sampled_voting_prefs = None
        self.posterior_mean_voting_prefs = None
        self.credible_interval_95_mean_voting_prefs = None
        self.num_groups_and_num_candidates = [None, None]

        self.turnout_adjusted_samples = None  # num_samples x num_precincts x r x (c-1)
        self.turnout_adjusted_sampled_voting_prefs = (
            None  # samps districtwide prefs,num_samples x r x c-1
        )
        self.turnout_adjusted_candidate_names = None  # candidate names with no-vote column omitted
        self.turnout_adjusted_posterior_mean_voting_prefs = None
        self.turnout_adjusted_credible_interval_95_mean_voting_prefs = None
        self.turnout_samples = None

    def fit(  # pylint: disable=too-many-branches
        self,
        group_fractions,
        votes_fractions,
        precinct_pops,
        demographic_group_names=None,
        candidate_names=None,
        target_accept=0.99,
        tune=1500,
        draw_samples=True,
        precinct_names=None,
        **other_sampling_args,
    ):
        """Fit the specified model using MCMC sampling
        Required arguments:
        group_fractions :   r x p (p =#precincts = num_precincts) matrix giving demographic
            information as the fraction of precinct_pop in the demographic group for each
            of p precincts and r demographic groups (sometimes denoted X)
        votes_fractions  :  c x p giving the fraction of each precinct_pop that votes
            for each of c candidates (sometimes denoted T)
        precinct_pops   :   Length-p array of ints giving size of each precinct population
                            of interest (e.g. voting population) (someteimes denoted N)
        Optional arguments:
        demographic_group_names  :  Names of the r demographic group of interest,
                                    where results are computed for the
                                    demographic group and its complement
        candidate_names          :  Name of the c candidates or voting outcomes of interest
        precinct_names          :   Length p vector giving the string names
                                    for each precinct.
        target_accept : float
            Strictly between zero and 1 (should be close to 1). Passed to pymc's
            sampling.sample
        tune : int
            Passed to pymc's sampling.sample
        draw_samples: bool, optional
            Default=True. Set to False to only set up the variable but not generate
            posterior samples (i.e. if you want to generate prior predictive samples only)
        other_sampling_args :
            For to pymc's sampling.sample
            https://docs.pymc.io/api/inference.html

        """
        # Additional params for hyperparameters
        # TODO: describe hyperparameters

        self.demographic_group_fractions = group_fractions
        self.votes_fractions = votes_fractions

        # check that precinct_pops are integers
        if not all(isinstance(p, (int, np.integer)) for p in precinct_pops):
            raise ValueError("all elements of precinct_pops must be integer-valued")
        self.precinct_pops = precinct_pops

        # check that group_fractions and vote_fractions sum to 1 in each precinct
        if not np.isclose(group_fractions.sum(axis=0), 1.0).all():
            raise ValueError("group_fractions should sum to 1 within each precinct")
        if not np.isclose(votes_fractions.sum(axis=0), 1.0).all():
            raise ValueError("votes_fractions should sum to 1 within each precinct")

        # check that group_fractions and vote_fractions are nonnegative
        if not (group_fractions >= 0).all():
            raise ValueError("group_fractions must be non-negative")
        if not (votes_fractions >= 0).all():
            raise ValueError("votes_fractions most be non-negative")

        # give demographic groups, candidates 1-indexed numbers as names if names are not specified
        if demographic_group_names is None:
            demographic_group_names = [str(i) for i in range(1, group_fractions.shape[0] + 1)]
        if candidate_names is None:
            candidate_names = [str(i) for i in range(1, votes_fractions.shape[0] + 1)]
        self.demographic_group_names = demographic_group_names
        self.candidate_names = candidate_names

        # Set precinct names
        if precinct_names is not None:  # pylint: disable=duplicate-code
            assert len(precinct_names) == len(precinct_pops)  # pylint: disable=duplicate-code
            if len(set(precinct_names)) != len(precinct_names):  # pylint: disable=duplicate-code
                warnings.warn(
                    "Precinct names are not unique. This may interfere with "
                    "passing precinct names to precinct_level_plot()."
                )
            self.precinct_names = np.array(precinct_names)

        self.num_groups_and_num_candidates = [
            group_fractions.shape[0],
            votes_fractions.shape[0],
        ]  # [r, c]

        check_dimensions_of_input(  # pylint: disable=duplicate-code
            group_fractions,  # pylint: disable=duplicate-code
            votes_fractions,  # pylint: disable=duplicate-code
            precinct_pops,  # pylint: disable=duplicate-code
            demographic_group_names,  # pylint: disable=duplicate-code
            candidate_names,
            self.num_groups_and_num_candidates,
        )

        if self.model_name == "multinomial-dirichlet":
            self.sim_model = ei_multinom_dirichlet(
                group_fractions,
                votes_fractions,
                precinct_pops,
                **self.additional_model_params,
            )

        elif self.model_name == "multinomial-dirichlet-modified":
            self.sim_model = ei_multinom_dirichlet_modified(
                group_fractions,
                votes_fractions,
                precinct_pops,
                **self.additional_model_params,
            )

        elif self.model_name == "greiner-quinn":
            self.sim_model = None

        else:
            raise ValueError(
                f"""{self.model_name} is not a supported model_name
            Currently supported: RxC models: 'multinomial-dirichlet-modified',
            'multinomial-dirichlet' """
            )

        if draw_samples:
            if self.model_name in [
                "multinomial-dirichlet-modified",
                "multinomial-dirichlet",
            ]:  # for models whose sampling is w/ pycm
                with self.sim_model:  # pylint: disable=not-context-manager
                    self.sim_trace = pm.sample(
                        target_accept=target_accept,
                        tune=tune,
                        nuts_sampler="numpyro",
                        **other_sampling_args,
                    )
            elif self.model_name == "greiner-quinn":
                self.sim_trace = pyei_greiner_quinn_sample(
                    group_fractions,
                    votes_fractions,
                    precinct_pops,
                    **other_sampling_args,
                )  #

            self.calculate_summary()

    def _calculate_turnout_adjusted_samples(self, non_candidate_names):
        """
        For each sample, calculate the voting support of each group for each candidate
        *as a fraction of all those who voted* (instead of as a fraction of all those
        included in the precinct population. This fn is only applicable when one of the
        c voting outcomes is a no-vote or abstain column. In this case, the total number
        of voters is unknown but samples from its distribution can be calculated)

        Parameters
        ----------
        non_candidate_names: list of str
            each a name of the column/ voting outcome that corresponds to not voting,
            if applicable. Each string in the list must be in candidate_names

        Notes
        -----
        Sets the values of:
            self.turnout_adjusted_candidate_names
            self.turnout_samples
            self.turnout_adjusted_samples
        """

        abstain_column_indices = []
        for non_candidate_name in non_candidate_names:
            if non_candidate_name not in self.candidate_names:
                raise ValueError(
                    f"non_candidate_names must be in candidate_names: {self.candidate_names}"
                )
            abstain_column_indices.append(self.candidate_names.index(non_candidate_name))

        non_adjusted_samples = np.transpose(
            self.sim_trace["posterior"]["b"].stack(all_draws=["chain", "draw"]).values,
            axes=(3, 0, 1, 2),
        )  # num_samples x num_precincts x r x c  # num_samples x num_precincts x r x c

        self.turnout_adjusted_candidate_names = [
            name for name in self.candidate_names if name not in non_candidate_names
        ]

        total_abstentions = non_adjusted_samples[:, :, :, abstain_column_indices].sum(
            axis=3
        )  # total fraction in all vote columnn num_samples x num_precincts x r

        self.turnout_samples = (
            1 - total_abstentions  # fraction that aren't in the no-vote column(s)
        ) * np.swapaxes(self.demographic_group_fractions * self.precinct_pops, 0, 1)

        turnout_adjusted_samples = np.delete(
            non_adjusted_samples, abstain_column_indices, axis=3
        )  # num_samples x num_precincts x r x c-1

        self.turnout_adjusted_samples = turnout_adjusted_samples / turnout_adjusted_samples.sum(
            axis=3, keepdims=True
        )  # num_samples x num_precincts x r x c-1

    def calculate_turnout_adjusted_summary(self, non_candidate_names):
        """
        Calculates districtwide samples, means, and credible intervals

        Parameters
        ----------
        non_candidate_names: list of str
            each a name of the column/ voting outcome that corresponds to not voting,
            if applicable. Each string in the list must be in candidate_names

        Notes
        -----
        Sets turnout_adjusted_voting_prefs, turnout_adjusted_posterior_mean_voting_prefs,
            turnout_adjusted_credible_interval_95_mean_voting_prefs
        """
        self._calculate_turnout_adjusted_samples(non_candidate_names)

        samples_converted_to_pops = (
            np.transpose(self.turnout_adjusted_samples, axes=(3, 0, 1, 2)) * self.turnout_samples
        )
        # (c-1) x num_samples x num_precincts x r x
        samples_of_votes_summed_across_district = samples_converted_to_pops.sum(
            axis=2
        )  # (c-1) x num_samples x r

        # obtain samples of the districtwide proportion of each demog. group voting for candidate
        self.turnout_adjusted_sampled_voting_prefs = np.transpose(
            samples_of_votes_summed_across_district / self.turnout_samples.sum(axis=1),
            axes=(1, 2, 0),
        )  # sampled voted prefs across precincts,  num_samples x r x c-1

        # # compute point estimates
        self.turnout_adjusted_posterior_mean_voting_prefs = (
            self.turnout_adjusted_sampled_voting_prefs.mean(axis=0)
        )  # r x (c -1)

        # # compute credible intervals
        percentiles = [2.5, 97.5]
        self.turnout_adjusted_credible_interval_95_mean_voting_prefs = np.zeros(
            (
                self.num_groups_and_num_candidates[0],
                self.num_groups_and_num_candidates[1] - 1,
                2,
            )
        )
        for row in range(self.num_groups_and_num_candidates[0]):
            for col in range(self.num_groups_and_num_candidates[1] - 1):
                self.turnout_adjusted_credible_interval_95_mean_voting_prefs[row][col][:] = (
                    np.percentile(
                        self.turnout_adjusted_sampled_voting_prefs[:, row, col], percentiles
                    )
                )

    def calculate_summary(self):
        """Calculate point estimates (post. means) and 95% equal-tailed credible intervals

        Sets
            self.sampled_voting_prefs
            self.posterior_mean_voting_prefs
            self.credible_interval_95_mean_voting_prefs
        """
        # multiply sample proportions by precinct pops for each group to get samples of
        # number of voters of the demographic group who voted for the candidate
        # in each precinct
        # This next messy line created to extract/reshape the InferenceData object to
        # match what was previously returned by self.sim_trace.get_values("b")
        # (needs to flatten out the chains dimension)
        b_values = np.transpose(
            self.sim_trace["posterior"]["b"].stack(all_draws=["chain", "draw"]).values,
            axes=(3, 0, 1, 2),
        )  # num_samples x num_precincts x r x c
        demographic_group_counts = np.transpose(
            self.demographic_group_fractions * self.precinct_pops
        )  # num_precincts x r
        samples_converted_to_pops = (
            np.transpose(b_values, axes=(3, 0, 1, 2)) * demographic_group_counts
        )
        # c x num_samples x num_precincts x r

        samples_of_votes_summed_across_district = samples_converted_to_pops.sum(
            axis=2
        )  # c x num_samples x r

        # obtain samples of the districtwide proportion of each demog. group voting for candidate
        self.sampled_voting_prefs = np.transpose(
            samples_of_votes_summed_across_district / demographic_group_counts.sum(axis=0),
            axes=(1, 2, 0),
        )  # sampled voted prefs across precincts,  num_samples x r x c

        # # compute point estimates
        self.posterior_mean_voting_prefs = self.sampled_voting_prefs.mean(axis=0)  # r x c

        # compute credible intervals
        percentiles = [2.5, 97.5]
        self.credible_interval_95_mean_voting_prefs = np.zeros(
            (
                self.num_groups_and_num_candidates[0],
                self.num_groups_and_num_candidates[1],
                2,
            )
        )
        for row in range(self.num_groups_and_num_candidates[0]):
            for col in range(self.num_groups_and_num_candidates[1]):
                self.credible_interval_95_mean_voting_prefs[row][col][:] = np.percentile(
                    self.sampled_voting_prefs[:, row, col], percentiles
                )

    def _calculate_margin(self, group, candidates, threshold=None, percentile=None):
        """
        Calculating the Candidate 1 - Candidate 2 margin among the given group.
        Calculate the percentile given a threshold, or vice versa. Exactly one of
        {percentile, threshold} must be None.

        Parameters:
        ----------
        group: str
            Demographic group in question
        candidates: list of str
            Length 2 vector of candidates upon which to calculate the margin
        threshold: float (optional)
            A specified level for the margin between the two candidates. If specified,
            use the threshold to calculate the percentile (% of samples with a larger margin)
        percentile: float (opetional)
            Between 0 and 100. Used to calculate the equal-tailed interval for the margin between
            the two candidates.

        Returns:
        --------
            threshold
            percentile
            samples
            group
            candidates
        """
        # TODO: document return values
        candidate_index_0 = self.candidate_names.index(candidates[0])
        candidate_index_1 = self.candidate_names.index(candidates[1])
        group_index = self.demographic_group_names.index(group)

        samples = (
            self.sampled_voting_prefs[:, group_index, candidate_index_0]
            - self.sampled_voting_prefs[:, group_index, candidate_index_1]
        )

        if percentile is None and threshold is not None:
            percentile = 100 * (samples > threshold).sum() / len(self.sampled_voting_prefs)
        elif threshold is None and percentile is not None:
            threshold = np.percentile(samples, 100 - percentile)
        else:
            raise ValueError(
                """Exactly one of threshold or percentile must be None.
            Set a threshold to calculate the associated percentile, or a percentile
            to calculate the associated threshold.
            """
            )
        return threshold, percentile, samples, group, candidates

    def margin_report(self, group, candidates, threshold=None, percentile=None, verbose=True):
        """
        For a given threshold, return the probability that the margin between
        the two candidates preferences in the given demographic group is greater than
        the threshold
        OR
        For a given confidence level, calculate the associated confidence interval
        of the difference between the two candidates preference among the group.
        Exactly one of {percentile, threshold} must be None.

        Parameters:
        -----------
        group: str
            Demographic group in question
        candidates: list of str
            Length 2 vector of candidates upon which to calculate the margin
        threshold: float (optional)
            A specified level for the margin between the two candidates. If specified,
            use the threshold to calculate the percentile (% of samples with a larger margin)
        percentile: float (opetional)
            Between 0 and 100. Used to calculate the equal-tailed interval for the margin between
            the two candidates.
        verbose: bool
            If true, print a report putting margin in context
        """
        return_interval = threshold is None

        if not all(candidate in self.candidate_names for candidate in candidates):
            raise ValueError(
                f"""candidate names must be in the list of candidate_names provided to fit():
                {self.candidate_names}"""
            )

        if group not in self.demographic_group_names:
            raise ValueError(
                f"""group name must be in the list of demographic_group_names
                provided to fit():
                {self.demographic_group_names}"""
            )

        if return_interval:
            lower_percentile = (100 - percentile) / 2
            upper_percentile = lower_percentile + percentile
            lower_threshold, _, _, group, candidates = self._calculate_margin(
                group, candidates, threshold, upper_percentile
            )
            upper_threshold, _, _, group, candidates = self._calculate_margin(
                group, candidates, threshold, lower_percentile
            )

            if verbose:
                print(
                    f"There is a {percentile}% probability that the difference between"
                    + f" {group}s' preferences for {candidates[0]} and {candidates[1]} is"
                    + f" between [{lower_threshold:.2f}, {upper_threshold:.2f}]."
                )
            return (lower_threshold, upper_threshold)
        else:
            threshold, percentile, _, group, candidates = self._calculate_margin(
                group, candidates, threshold, percentile
            )
            if verbose:
                print(
                    f"There is a {percentile:.1f}% probability that the difference between"
                    + f" {group}s' preferences for {candidates[0]} and {candidates[1]}"
                    + f" is more than {threshold:.2f}."
                )
            return percentile

    def _calculate_polarization(self, groups, candidate, threshold=None, percentile=None):
        """
        Calculate percentile given a threshold, or vice versa.
        Exactly one of {percentile, threshold} must be None.
        Parameters:
        -----------
        groups: array-like
            Length 2 vector of demographic groups from which to calculate polarization
        candidate: string
            Candidate for which to calculate polarization
        threshold: float (optional)
            A specified level of difference in support for the candidate
            between one group and the other. If specified, use the threshold
            to calculate the percentile (exactly one of threshold and percentile
            must be None)
        percentile: float (optional)
            Between 0 and 100. Used to calculate the equal-tailed interval
            for the polarization. At least one of threshold and percentile
            must be None
        """

        candidate_index = self.candidate_names.index(candidate)
        group_index_0 = self.demographic_group_names.index(groups[0])
        group_index_1 = self.demographic_group_names.index(groups[1])

        samples = (
            self.sampled_voting_prefs[:, group_index_0, candidate_index]
            - self.sampled_voting_prefs[:, group_index_1, candidate_index]
        )

        if percentile is None and threshold is not None:
            percentile = 100 * (samples > threshold).sum() / len(self.sampled_voting_prefs)
        elif threshold is None and percentile is not None:
            threshold = np.percentile(samples, 100 - percentile)
        else:
            raise ValueError(
                """Exactly one of threshold or percentile must be None.
            Set a threshold to calculate the associated percentile, or a percentile
            to calculate the associated threshold.
            """
            )
        return threshold, percentile, samples, groups, candidate

    def polarization_report(self, groups, candidate, threshold=None, percentile=None, verbose=True):
        """
        For a given threshold, return the probability that the difference between
        the two demographic groups' preferences for the candidate is greater than
        the threshold
        OR
        For a given confidence level, calculate the associated confidence interval
        of the difference between the two groups' preferences.
        Exactly one of {percentile, threshold} must be None.
        Parameters:
        -----------
        groups:
            Length 2 vector of demographic groups from which to calculate polarization
        candidate: string
            Candidate for which to calculate polarization
        threshold: float (optional)
            A specified level of difference in support for the candidate
            between one group and the other. If specified, use the threshold
            to calculate the percentile (exactly one of threshold and percentile
            must be None)
        percentile: float (optional)
            Between 0 and 100. Used to calculate the equal-tailed interval
            for the polarization. At least one of threshold and percentile
            must be None
        verbose: bool
            If true, print a report putting polarization in context

        """
        return_interval = threshold is None

        if not all(group in self.demographic_group_names for group in groups):
            raise ValueError(
                f"""Elements of group_names must be in the list of demographic_group_names
                provided to fit():
                {self.demographic_group_names}"""
            )

        if candidate not in self.candidate_names:
            raise ValueError(
                f"""candidate_name must be in the list of candidate_names provided to fit():
                {self.candidate_names}"""
            )

        if return_interval:
            lower_percentile = (100 - percentile) / 2
            upper_percentile = lower_percentile + percentile
            lower_threshold, _, _, groups, candidate = self._calculate_polarization(
                groups, candidate, threshold, upper_percentile
            )
            upper_threshold, _, _, groups, candidate = self._calculate_polarization(
                groups, candidate, threshold, lower_percentile
            )

            if verbose:
                print(
                    f"There is a {percentile}% probability that the difference between"
                    + f" the groups' preferences for {candidate} ({groups[0]} - {groups[1]}) is"
                    + f" between [{lower_threshold:.2f}, {upper_threshold:.2f}]."
                )
            return (lower_threshold, upper_threshold)
        else:
            threshold, percentile, _, groups, candidate = self._calculate_polarization(
                groups, candidate, threshold, percentile
            )
            if verbose:
                print(
                    f"There is a {percentile:.1f}% probability that the difference between"
                    + f" the groups' preferences for {candidate} ({groups[0]} - {groups[1]}) "
                    + f" is more than {threshold:.2f}."
                )
            return percentile

    def summary(self, non_candidate_names=None):
        """Return a summary string for the ei results

        Parameters:
        -----------
        non_candidate_names: list of str (optional)
            Each a string in self.candidate_names() that corresponds to the name of a "no-vote"
            or abstain column, if applicable. If passed, the summary will be estimates of
            support for candidates AMONG those who were estimated to not be in the
            specified abstain column
        """
        # TODO: probably format this as a table
        summary_str = """
            Computed from the raw b_ samples by multiplying by group populations and then
            getting the proportion of the total pop
            (total pop=summed across all districts):
            """
        if non_candidate_names is not None:
            self.calculate_turnout_adjusted_summary(non_candidate_names)
            candidate_names_for_summary = self.turnout_adjusted_candidate_names
            posterior_means = self.turnout_adjusted_posterior_mean_voting_prefs
            credible_intervals = self.turnout_adjusted_credible_interval_95_mean_voting_prefs
        else:
            candidate_names_for_summary = self.candidate_names
            posterior_means = self.posterior_mean_voting_prefs
            credible_intervals = self.credible_interval_95_mean_voting_prefs

        for row in range(self.num_groups_and_num_candidates[0]):
            for col, candidate_name in enumerate(candidate_names_for_summary):
                summ = f"""The posterior mean for the district-level voting preference of
                {self.demographic_group_names[row]} for {candidate_name} is
                {posterior_means[row][col]:.3f}
                95% equal-tailed credible interval:  {credible_intervals[row][col]}
                """
                summary_str += summ
        return summary_str

    def precinct_level_estimates(self, non_candidate_names=None):
        """Returns precinct-level posterior means and credible intervals

        Parameters:
        ----------
        non_candidate_names: list of str
            Optional. If specified, this will give the names of column to be
            treated as no-vote columns, and the precinct-level estimates
            will be computed AMONG those who were estimated to have voted
        Returns:
        --------
            precinct_posterior_means: num_precincts x r x c
            precinct_credible_intervals: num_precincts x r x c x 2
        """
        if non_candidate_names is not None:
            precinct_level_samples = self.turnout_adjusted_samples
        else:
            precinct_level_samples = np.transpose(
                self.sim_trace["posterior"]["b"].stack(all_draws=["chain", "draw"]).values,
                axes=(3, 0, 1, 2),
            )  # num_samples x num_precincts x r x c # num_samples x num_precincts x r x c
        _, _, r, c = precinct_level_samples.shape
        precinct_posterior_means = precinct_level_samples.mean(axis=0)
        precinct_credible_intervals = np.ones(
            (
                len(self.precinct_pops),
                r,
                c,
                2,
            )
        )
        percentiles = [2.5, 97.5]

        for row in range(r):
            for col in range(c):
                precinct_credible_intervals[:, row, col, :] = np.percentile(
                    precinct_level_samples[:, :, row, col], percentiles, axis=0
                ).T

        return (precinct_posterior_means, precinct_credible_intervals)

    def candidate_of_choice_report(self, verbose=True, non_candidate_names=None):
        """For each group, look at differences in preference within that group
        Parameters:
        -----------
        verbose: boolean (optional)
            If true, print a report putting the candidate preference rate in context
            If false, do not print report.
            In either case, return the candidate preference dictionary
        non_candidate_names: list of str (optional)
            A list of strings giving the names of voting outcomes that should not be
            considered candidates for the purposes of calculating the candidate of choice.
            For example, if there is an 'Abstain' column, we want may want to to include
            'Abstain' in our list of non_candidate_names so that we only consider candidates
            of choice to be actual candidates.

        Returns:
        -------
        candidate_preference_rate_dict: dict
            keys are of the form (demographic group name, candidate name)
            Values are fraction of the samples in which the support of that group for that
            candidate was higher than for any other candidate
        """

        candidate_preference_rate_dict = {}
        if non_candidate_names is None:
            non_candidate_names = []
        non_cand_idxs = [self.candidate_names.index(n) for n in non_candidate_names]
        cand_names = [c for c in self.candidate_names if c not in non_candidate_names]
        sampled_voting_prefs = np.delete(self.sampled_voting_prefs, non_cand_idxs, axis=2)

        for row in range(self.num_groups_and_num_candidates[0]):
            if verbose:
                print(self.demographic_group_names[row])
            for candidate_idx, name in enumerate(cand_names):
                frac = (
                    np.argmax(sampled_voting_prefs[:, row, :], axis=1) == candidate_idx
                ).sum() / sampled_voting_prefs.shape[0]
                if verbose:
                    print(
                        f"     - In {round(frac*100,3)} percent of samples, the district-level "
                        f"vote preference of \n"
                        f"       {self.demographic_group_names[row]} for "
                        f"{name} "
                        f"was higher than for any other candidate."
                    )
                candidate_preference_rate_dict[(self.demographic_group_names[row], name)] = frac
        return candidate_preference_rate_dict

    def candidate_of_choice_polarization_report(self, verbose=True, non_candidate_names=None):
        """For each pair of groups, look at differences in preferences
        between those groups

        Parameters:
        -----------
        verbose: boolean (optional)
            If true, print a report putting the candidate of choice polarization rate context
            If false, do not pritn report.
            In either case, return the candidate difference rate dictionary
        non_candidate_names: list of str (optional)
            A list of strings giving the names of voting outcomes that should not be
            considered candidates for the purposes of calculating the candidate of choice.
            For example, if there is an 'Abstain' column, we want may want to to include
            'Abstain' in our list of non_candidate_names so that we only consider candidates
            of choice to be actual candidates.

        Returns:
        --------
        candidate_differ_rate_dict: dict
            Keys are of the form (group1, group2), values are the fraction of samples
            for which those two groups had a different candidate of choice

        Notes:
        ------
        For each pair of groups, this function reports the fraction samples for which
        the `preferred candidate` of one group (as measured by: who is the candidate supported
        by the plurality within that group according to the sampled distric-level support value)
        is different from the `preferred candidate` of the others group
        """

        candidate_differ_rate_dict = {}
        if non_candidate_names is None:
            non_candidate_names = []
        non_cand_idxs = [self.candidate_names.index(n) for n in non_candidate_names]
        sampled_voting_prefs = np.delete(self.sampled_voting_prefs, non_cand_idxs, axis=2)

        for dem1 in range(self.num_groups_and_num_candidates[0]):
            for dem2 in range(dem1):
                differ_frac = (
                    np.argmax(sampled_voting_prefs[:, dem1, :], axis=1)
                    != np.argmax(sampled_voting_prefs[:, dem2, :], axis=1)
                ).sum() / sampled_voting_prefs.shape[0]
                if verbose:
                    print(
                        f"In {round(differ_frac*100,3)} percent of samples, the district-level "
                        f"candidates of choice for {self.demographic_group_names[dem1]} and "
                        f"{self.demographic_group_names[dem2]} voters differ."
                    )
                candidate_differ_rate_dict[
                    (
                        self.demographic_group_names[dem1],
                        self.demographic_group_names[dem2],
                    )
                ] = differ_frac
                candidate_differ_rate_dict[
                    (
                        self.demographic_group_names[dem2],
                        self.demographic_group_names[dem1],
                    )
                ] = differ_frac
        return candidate_differ_rate_dict

    def plot(self, non_candidate_names=None):
        """Plot with no arguments returns the kde plots, with one plot for each candidate

        Parameters:
        non_candidate_names: list of str
            each a name of the column/ voting outcome that corresponds to not voting,
            if applicable. Each string in the list must be in candidate_names
        """
        return self.plot_kdes(
            plot_by="candidate", non_candidate_names=non_candidate_names, axes=None
        )

    def plot_boxplots(self, plot_by="candidate", non_candidate_names=None, axes=None):
        """Plot boxplots of voting prefs (one boxplot for each candidate)

        Parameters:
        -----------
        plot_by: {'candidate', 'group'}, optional
            If candidate, make one plot for each candidate. If group, make
            one subplot for each group. (Default: 'candidate')
        axes: list of Matplotlib axis objects, optional
            Typically subplots within the same figure. Length c if plot_by = 'candidate',
            length r if plot_by = 'group'
        """
        if non_candidate_names is None:
            voting_prefs = self.sampled_voting_prefs
            candidate_names = self.candidate_names
        else:  # turnout adjusted samples, names without no-vote column
            self.calculate_turnout_adjusted_summary(non_candidate_names)
            voting_prefs = self.turnout_adjusted_sampled_voting_prefs
            candidate_names = self.turnout_adjusted_candidate_names

        return plot_boxplots(
            voting_prefs,
            self.demographic_group_names,
            candidate_names,
            plot_by=plot_by,
            axes=axes,
        )

    def plot_kdes(self, plot_by="candidate", non_candidate_names=None, axes=None):
        """Kernel density plots of voting preference, plots grouped by candidate or group

        Parameters:
        -----------
        plot_by: {'candidate', 'group'}, optional
            If candidate, make one plot for each candidate. If group, make
            one subplot for each group (Default: 'candidate')
        axes: list of Matplotlib axis objects, optional
            Typically subplots within the same figure. Length c if plot_by = 'candidate',
            length r if plot_by = 'group'
        """
        if non_candidate_names is None:
            voting_prefs = self.sampled_voting_prefs
            candidate_names = self.candidate_names
        else:  # turnout adjusted samples, names without no-vote column
            self.calculate_turnout_adjusted_summary(non_candidate_names)
            voting_prefs = self.turnout_adjusted_sampled_voting_prefs
            candidate_names = self.turnout_adjusted_candidate_names

        return plot_kdes(
            voting_prefs,
            self.demographic_group_names,
            candidate_names,
            plot_by=plot_by,
            axes=axes,
        )

    def plot_margin_kde(
        self,
        group,
        candidates,
        threshold=None,
        percentile=None,
        show_threshold=False,
        ax=None,
    ):
        """
        Plot kde of the margin between two candidates among the given demographic group.

        Parameters:
        ----------
        group: str
            Demographic group in question
        candidates: list of str
            Length 2 vector of candidates upon which to calculate the margin
        threshold: float (optional)
            A specified level for the margin between the two candidates. If specified,
            use the threshold to calculate the percentile (% of samples with a larger margin)
        percentile: float (opetional)
            Between 0 and 100. Used to calculate the equal-tailed interval for the margin between
            the two candidates.
        show_threshold: bool
            Show threshold in the plot.
        """
        return_interval = threshold is None
        if return_interval:
            lower_percentile = (100 - percentile) / 2
            upper_percentile = lower_percentile + percentile
            lower_threshold, _, samples, group, candidates = self._calculate_margin(
                group, candidates, threshold, upper_percentile
            )
            upper_threshold, _, samples, group, candidates = self._calculate_margin(
                group, candidates, threshold, lower_percentile
            )
            thresholds = [lower_threshold, upper_threshold]
        else:
            threshold, percentile, samples, group, candidates = self._calculate_margin(
                group, candidates, threshold, percentile
            )
            thresholds = [threshold]

        return plot_margin_kde(
            group, candidates, samples, thresholds, percentile, show_threshold, ax
        )

    def plot_polarization_kde(
        self,
        groups,
        candidate,
        threshold=None,
        percentile=None,
        show_threshold=False,
        ax=None,
        color="steelblue",
    ):
        """Plot kde of differences between voting preferences

        Parameters:
        -----------
        groups : list of strings
            Names of the demographic groups
        candidate: str
            Name of the candidate for whom the polarization is calculated
        threshold: float (optional)
            A specified level of difference in support for the candidate
            between one group and the other. If specified, use the threshold
            to calculate the percentile (exactly one of threshold and percentile
            must be None)
        percentile: float (optional)
            Between 0 and 100. Used to calculate the equal-tailed interval
            for the polarization. At least one of threshold and percentile
            must be None
        show_threshold: bool
        ax : matplotlib Axis object
        color : str (optional)
            Specifies a color for matplotlib to be used in the histogram/kde.
            default="steelblue"

        Returns:
        --------
        matplotlib axis object
        """
        return_interval = threshold is None

        if return_interval:
            lower_percentile = (100 - percentile) / 2
            upper_percentile = lower_percentile + percentile
            lower_threshold, _, samples, groups, candidate = self._calculate_polarization(
                groups, candidate, threshold, upper_percentile
            )
            upper_threshold, _, samples, groups, candidate = self._calculate_polarization(
                groups, candidate, threshold, lower_percentile
            )
            thresholds = [lower_threshold, upper_threshold]
        else:
            threshold, percentile, samples, groups, candidate = self._calculate_polarization(
                groups, candidate, threshold, percentile
            )
            thresholds = [threshold]

        return plot_polarization_kde(
            samples,
            thresholds,
            percentile,
            groups,
            candidate,
            show_threshold,
            ax,
            color=color,
        )

    def plot_intervals_by_precinct(self, group_name, candidate_name):
        """Plot of credible intervals for all precincts, for specified group and candidate

        Parameters:
        -----------
        group_name : str
            Group for which to plot intervals. Should be in demographic_group_names
        candiate_name : str
            Candidate for which to plot intervals. Should be in candidate_names
        """
        if group_name not in self.demographic_group_names:
            raise ValueError(
                f"""group_name must be in the list of demographic_group_names provided to fit():
                {self.demographic_group_names}"""
            )

        if candidate_name not in self.candidate_names:
            raise ValueError(
                f"""candidate_name must be in the list of candidate_names provided to fit():
                {self.candidate_names}"""
            )

        group_index = self.demographic_group_names.index(group_name)
        candidate_index = self.candidate_names.index(candidate_name)

        point_estimates_all, intervals_all = self.precinct_level_estimates()
        point_estimates = point_estimates_all[:, group_index, candidate_index]
        intervals = intervals_all[:, group_index, candidate_index, :]

        return plot_intervals_all_precincts(
            point_estimates,
            intervals,
            candidate_name,
            self.precinct_names,
            group_name,  # TODO: _group_names_for_display?
            ax=None,
            show_all_precincts=False,
        )

    def precinct_level_plot(
        self,
        candidate,
        groups=None,
        alpha=1,
        ax=None,
        show_all_precincts=False,
        precinct_names=None,
        plot_as_histograms=False,
    ):
        """
        Optional arguments:
        candidate           : str
                                The candidate whose support we're examining
        groups              : list of str
                                The groups whose support we're examining
        alpha               : float
                                The opacity of the ridgeplots' fill color
        ax                  :  matplotlib axes object
        show_all_precincts  :  If True, then it will show all ridge plots
                               (even if there are more than 50)
        precinct_names      :  Labels for each precinct (if not supplied, by
                               default we label each precinct with an integer
                               label, 1 to n)
        plot_as_histograms : bool, optional. Default is false. If true, plot
                                with histograms instead of kdes
        """
        precinct_level_samples = np.transpose(
            self.sim_trace["posterior"]["b"].stack(all_draws=["chain", "draw"]).values,
            axes=(3, 0, 1, 2),
        )  # num_samples x num_precincts x r x c  # num_samples x num_precincts x r x c
        groups = self.demographic_group_names if groups is None else groups
        candidate_idx = self.candidate_names.index(candidate)
        voting_prefs = []
        for group in groups:
            group_idx = self.demographic_group_names.index(group)
            voting_prefs.append(precinct_level_samples[:, :, group_idx, candidate_idx])
        return plot_precincts(
            voting_prefs,
            group_names=groups,
            candidate=candidate,
            alpha=alpha,
            precinct_labels=precinct_names,  # pylint: disable=duplicate-code
            show_all_precincts=show_all_precincts,
            plot_as_histograms=plot_as_histograms,
            ax=ax,
        )



"""Test two by two ecological inference."""

import pytest
import numpy as np
import scipy.stats as st
from scipy.special import logsumexp

from pyei import two_by_two
from pyei import data
from pyei.two_by_two import TwoByTwoEI


@pytest.fixture(scope="session")
def example_two_by_two_data():
    """load santa clara data to test two by two ei and plots"""  #
    sc_data = data.Datasets.Santa_Clara.to_dataframe()
    group_fractions = np.array(sc_data["pct_e_asian_vote"])
    votes_fractions = np.array(sc_data["pct_for_hardy2"])  #
    precinct_pops = np.array(sc_data["total2"])
    demographic_group_name = "e_asian"
    candidate_name = "Hardy"
    precinct_names = sc_data["precinct"]
    return {  #
        "group_fractions": group_fractions,
        "votes_fractions": votes_fractions,
        "precint_pops": precinct_pops,  #
        "demographic_group_name": demographic_group_name,
        "candidate_name": candidate_name,
        "precinct_names": precinct_names,
    }  #


@pytest.fixture(scope="session")
def example_two_by_two_ei(example_two_by_two_data):  # pylint: disable=redefined-outer-name
    """run example two by two ei method - can use to test plotting"""
    ei_ex = TwoByTwoEI(model_name="king99_pareto_modification", pareto_scale=8, pareto_shape=2)
    ei_ex.fit(  #
        example_two_by_two_data["group_fractions"],
        example_two_by_two_data["votes_fractions"],
        example_two_by_two_data["precint_pops"],  #
        demographic_group_name=example_two_by_two_data["demographic_group_name"],
        candidate_name=example_two_by_two_data["candidate_name"],
        precinct_names=example_two_by_two_data["precinct_names"],  #
        draws=100,
        tune=100,
    )
    return ei_ex


def generate_kwargs_for_log_binom_sum():
    """Randomly generate inputs for log_binom_sum.

    Note that these should probably _not_ be random eventually, but should deterministically
    pass. For now, this will give us more confidence that the implementation matches scipy for
    a range of numeric inputs.
    """
    lower = np.random.poisson(10)
    upper = lower + np.random.poisson(100)
    obs_vote = np.random.randint(lower, upper)
    kwargs = {
        "n0_curr": np.random.randint(obs_vote, upper),
        "n1_curr": np.random.randint(obs_vote, upper),
        "b_1_curr": np.random.rand(),
        "b_2_curr": np.random.rand(),
        "prev": np.random.randn(),
    }
    kwargs["lower"] = lower
    kwargs["upper"] = upper
    kwargs["obs_vote"] = obs_vote
    return kwargs


def log_binom_sum_in_scipy(lower, upper, obs_vote, n0_curr, n1_curr, b_1_curr, b_2_curr, prev):
    """Reimplement theano logic in scipy to make sure it matches."""
    votes_withing_group_count = np.arange(lower, upper)
    return (
        logsumexp(
            st.binom(n0_curr, b_1_curr).logpmf(votes_withing_group_count)
            + st.binom(n1_curr, b_2_curr).logpmf(obs_vote - votes_withing_group_count)
        )
        + prev
    )


def test_log_binom_sum():
    kwargs = generate_kwargs_for_log_binom_sum()
    np.testing.assert_almost_equal(
        two_by_two.log_binom_sum(**kwargs).eval(), log_binom_sum_in_scipy(**kwargs), decimal=4
    )


def test_binom_conv_log_p():
    # Randomly generate a dataset of 5 precincts
    sample_data = []
    for _ in range(5):
        kwargs = generate_kwargs_for_log_binom_sum()
        kwargs.pop("prev")
        sample_data.append(kwargs)

    # Unpack and repack for binom_conv_log_p
    b_1, b_2, n_0, n_1, upper, lower, obs_votes = list(
        zip(
            *[
                (
                    v["b_1_curr"],
                    v["b_2_curr"],
                    v["n0_curr"],
                    v["n1_curr"],
                    v["upper"],
                    v["lower"],
                    v["obs_vote"],
                )
                for v in sample_data
            ]
        )
    )

    theano_result = two_by_two.binom_conv_log_p(b_1, b_2, n_0, n_1, upper, lower, obs_votes).eval()

    # Now compute the result using scipy
    prev = np.array([0])
    for kwargs in sample_data:
        kwargs["prev"] = prev
        prev = log_binom_sum_in_scipy(**kwargs)

    np.testing.assert_allclose(theano_result, prev)


def test_polarization_report(example_two_by_two_ei):  # pylint: disable=redefined-outer-name
    prob_20 = example_two_by_two_ei.polarization_report(threshold=0.2)
    prob_40 = example_two_by_two_ei.polarization_report(threshold=0.4)
    thresh_95_range = example_two_by_two_ei.polarization_report(percentile=95)
    thresh_90_range = example_two_by_two_ei.polarization_report(percentile=90)

    assert prob_20 >= prob_40
    assert thresh_95_range[1] > thresh_95_range[0]
    assert thresh_95_range[1] - thresh_95_range[0] >= thresh_90_range[1] - thresh_90_range[0]
